# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase, ET


class UserActivity(ElementBase):

    name = 'activity'
    namespace = 'http://jabber.org/protocol/activity'
    plugin_attrib = 'activity'
    interfaces = {'value', 'text'}
    sub_interfaces = {'text'}
    general = {'doing_chores', 'drinking', 'eating', 'exercising',
               'grooming', 'having_appointment', 'inactive', 'relaxing',
               'talking', 'traveling', 'undefined', 'working'}
    specific = {'at_the_spa', 'brushing_teeth', 'buying_groceries',
                'cleaning', 'coding', 'commuting', 'cooking', 'cycling',
                'dancing', 'day_off', 'doing_maintenance',
                'doing_the_dishes', 'doing_the_laundry', 'driving',
                'fishing', 'gaming', 'gardening', 'getting_a_haircut',
                'going_out', 'hanging_out', 'having_a_beer',
                'having_a_snack', 'having_breakfast', 'having_coffee',
                'having_dinner', 'having_lunch', 'having_tea', 'hiding',
                'hiking', 'in_a_car', 'in_a_meeting', 'in_real_life',
                'jogging', 'on_a_bus', 'on_a_plane', 'on_a_train',
                'on_a_trip', 'on_the_phone', 'on_vacation',
                'on_video_phone', 'other', 'partying', 'playing_sports',
                'praying', 'reading', 'rehearsing', 'running',
                'running_an_errand', 'scheduled_holiday', 'shaving',
                'shopping', 'skiing', 'sleeping', 'smoking',
                'socializing', 'studying', 'sunbathing', 'swimming',
                'taking_a_bath', 'taking_a_shower', 'thinking',
                'walking', 'walking_the_dog', 'watching_a_movie',
                'watching_tv', 'working_out', 'writing'}

    def set_value(self, value):
        self.del_value()
        general = value
        specific = None
        if isinstance(value, tuple) or isinstance(value, list):
            general = value[0]
            specific = value[1]

        if general in self.general:
            gen_xml = ET.Element('{%s}%s' % (self.namespace, general))
            if specific:
                spec_xml = ET.Element('{%s}%s' % (self.namespace, specific))
                if specific in self.specific:
                    gen_xml.append(spec_xml)
                else:
                    raise ValueError('Unknown specific activity')
            self.xml.append(gen_xml)
        else:
            raise ValueError('Unknown general activity')

    def get_value(self):
        general = None
        specific = None
        gen_xml = None
        for child in self.xml:
            if child.tag.startswith('{%s}' % self.namespace):
                elem_name = child.tag.split('}')[-1]
                if elem_name in self.general:
                    general = elem_name
                    gen_xml = child
        if gen_xml is not None:
            for child in gen_xml:
                if child.tag.startswith('{%s}' % self.namespace):
                    elem_name = child.tag.split('}')[-1]
                    if elem_name in self.specific:
                        specific = elem_name
        return (general, specific)

    def del_value(self):
        curr_value = self.get_value()
        if curr_value[0]:
            self._set_sub_text(curr_value[0], '', keep=False)



import unittest
from slixmpp import Message
from slixmpp.jid import JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0482 import stanza
from slixmpp.plugins.xep_0482.stanza import External, Jingle
from slixmpp.xmlstream import register_stanza_plugin


class TestCallInviteStanza(SlixTest):

    def setUp(self):
        stanza.register_plugins()

    def test_invite(self):
        """Test that the element is created correctly."""
        msg = Message()
        msg['call-invite']['video'] = True
        jingle = Jingle()
        jingle['sid'] = 'toto'
        jingle['jid'] = JID('toto@example.com/m')
        external = External()
        external['uri'] = "https://example.com/call"
        msg['call-invite'].append(jingle)
        msg['call-invite'].append(external)

        self.check(msg, """
<message>
  <invite xmlns="urn:xmpp:call-invites:0" video="true">
    <jingle sid="toto" jid="toto@example.com/m" />
    <external uri="https://example.com/call" />
  </invite>
</message>
        """)

        self.assertEqual(
            msg['call-invite'].get_methods(),
            ([jingle], [external]),
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestCallInviteStanza)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp
from slixmpp.exceptions import XMPPError

log = logging.getLogger(__name__)


class MAM(slixmpp.ClientXMPP):

    """
    A basic client fetching mam archive messages
    """

    def __init__(self, jid, password, remote_jid, start):
        slixmpp.ClientXMPP.__init__(self, jid, password)
        self.remote_jid = remote_jid
        self.start_date = start

        self.add_event_handler("session_start", self.start)

    async def start(self, *args):
        """
        Fetch mam results for the specified JID.
        Use RSM to paginate the results.
        """
        results = self.plugin['xep_0313'].retrieve(jid=self.remote_jid, iterator=True, rsm={'max': 10}, start=self.start_date)
        page = 1
        async for rsm in results:
            print('Page %d' % page)
            for msg in rsm['mam']['results']:
                forwarded = msg['mam_result']['forwarded']
                timestamp = forwarded['delay']['stamp']
                message = forwarded['stanza']
                print('[%s] %s: %s' % (timestamp, message['from'], message['body']))
            page += 1
        self.disconnect()

if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()
    parser.add_argument("-q","--quiet", help="set logging to ERROR",
                        action="store_const",
                        dest="loglevel",
                        const=logging.ERROR,
                        default=logging.INFO)
    parser.add_argument("-d","--debug", help="set logging to DEBUG",
                        action="store_const",
                        dest="loglevel",
                        const=logging.DEBUG,
                        default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    # Other options
    parser.add_argument("-r", "--remote-jid", dest="remote_jid",
                        help="Remote JID")
    parser.add_argument("--start", help="Start date", default='2017-09-20T12:00:00Z')

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")
    if args.remote_jid is None:
        args.remote_jid = input("Remote JID: ")
    if args.start is None:
        args.start = input("Start time: ")

    xmpp = MAM(args.jid, args.password, args.remote_jid, args.start)
    xmpp.register_plugin('xep_0313')

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_until_complete(xmpp.disconnected)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2015 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import asyncio
import logging
from uuid import uuid4

from slixmpp.plugins import BasePlugin
from slixmpp import Iq, Message
from slixmpp.jid import JID
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0070 import stanza, Confirm


log = logging.getLogger(__name__)


class XEP_0070(BasePlugin):

    """
    XEP-0070 Verifying HTTP Requests via XMPP
    """

    name = 'xep_0070'
    description = 'XEP-0070: Verifying HTTP Requests via XMPP'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, Confirm)
        register_stanza_plugin(Message, Confirm)

        self.xmpp.register_handler(
            Callback('Confirm',
                 StanzaPath('iq@type=get/confirm'),
                 self._handle_iq_confirm))

        self.xmpp.register_handler(
            Callback('Confirm',
                 StanzaPath('message/confirm'),
                 self._handle_message_confirm))

    def plugin_end(self):
        self.xmpp.remove_handler('Confirm')
        self.xmpp['xep_0030'].del_feature(feature='http://jabber.org/protocol/http-auth')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('http://jabber.org/protocol/http-auth')

    def ask_confirm(self, jid, id, url, method, *, ifrom=None, message=None):
        jid = JID(jid)
        if jid.resource:
            stanza = self.xmpp.Iq()
            stanza['type'] = 'get'
        else:
            stanza = self.xmpp.Message()
            stanza['thread'] = uuid4().hex
        stanza['from'] = ifrom
        stanza['to'] = jid
        stanza['confirm']['id'] = id
        stanza['confirm']['url'] = url
        stanza['confirm']['method'] = method
        if not jid.resource:
            if message is not None:
                stanza['body'] = message.format(id=id, url=url, method=method)
            stanza.send()
            fut = asyncio.Future()
            fut.set_result(stanza)
            return fut
        else:
            return stanza.send()

    def _handle_iq_confirm(self, iq):
        self.xmpp.event('http_confirm_iq', iq)
        self.xmpp.event('http_confirm', iq)

    def _handle_message_confirm(self, message):
        self.xmpp.event('http_confirm_message', message)
        self.xmpp.event('http_confirm', message)



import unittest
from slixmpp.test import SlixTest


class TestJabberSearch(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component",
            plugin_config={
                "xep_0055": {
                    "form_fields": {"first", "last"},
                    "form_instructions": "INSTRUCTIONS",
                    "form_title": "User Directory Search",
                }
            },
            jid="characters.shakespeare.lit",
            plugins={"xep_0055"}
        )
        self.xmpp["xep_0055"].api.register(get_results, "search_query")
        self.xmpp["xep_0055"].api.register(get_results, "search_query")

    def testRequestingSearchFields(self):
        self.recv(
            """
            <iq type='get'
                from='juliet@capulet.com/balcony'
                to='characters.shakespeare.lit'
                id='search3'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'/>
            </iq>
            """
        )
        self.send(
            """
            <iq type='result'
                from='characters.shakespeare.lit'
                to='juliet@capulet.com/balcony'
                id='search3'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='form'>
                  <title>User Directory Search</title>
                  <instructions>INSTRUCTIONS</instructions>
                  <field type='hidden'
                         var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <field var='first'/>
                  <field var='last'/>
                </x>
              </query>
            </iq>
            """,
            use_values=False,
        )

    def testSearchResult(self):
        self.recv(
            """
            <iq type='get'
                from='juliet@capulet.com/balcony'
                to='characters.shakespeare.lit'
                id='search2'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field type='hidden' var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <field var='last'>
                    <value>Montague</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )
        self.send(
            """
            <iq type='result'
                from='characters.shakespeare.lit'
                to='juliet@capulet.com/balcony'
                id='search2'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='result'>
                  <field type='hidden' var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <reported>
                    <field var='first' label='Given Name' />
                    <field var='last' label='Family Name' />
                  </reported>
                  <item>
                    <field var='first'><value>Benvolio</value></field>
                    <field var='last'><value>Montague</value></field>
                  </item>
                </x>
              </query>
            </iq>
            """,
            use_values=False,  # TypeError: element indices must be integers without that
        )

    def testSearchNoResult(self):
        self.xmpp["xep_0055"].api.register(get_results, "search_query")
        self.recv(
            """
            <iq type='get'
                from='juliet@capulet.com/balcony'
                to='characters.shakespeare.lit'
                id='search2'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field type='hidden' var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <field var='last'>
                    <value>Capulet</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )
        self.send(
            """
            <iq type='result'
                from='characters.shakespeare.lit'
                to='juliet@capulet.com/balcony'
                id='search2'
                xml:lang='en'>
              <query xmlns='jabber:iq:search'>
                <x xmlns='jabber:x:data' type='result'>
                  <field type='hidden' var='FORM_TYPE'>
                    <value>jabber:iq:search</value>
                  </field>
                  <reported>
                    <field var='first' label='Given Name' />
                    <field var='last' label='Family Name' />
                  </reported>
                </x>
              </query>
            </iq>
            """,
            use_values=False,  # TypeError: element indices must be integers without that
        )

async def get_results(jid, node, ifrom, iq):
    reply = iq.reply()
    form = reply["search"]["form"]
    form["type"] = "result"

    form.add_reported("first", label="Given Name")
    form.add_reported("last", label="Family Name")

    d = iq["search"]["form"].get_values()

    if d["last"] == "Montague":
        form.add_item({"first": "Benvolio", "last": "Montague"})

    return reply


suite = unittest.TestLoader().loadTestsFromTestCase(TestJabberSearch)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dalek
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Invite(ElementBase):

    """
    XMPP allows for an agent in an MUC room to directly invite another
    user to join the chat room (as opposed to a mediated invitation
    done through the server).

    Example invite stanza:

    ::

      <message from='crone1@shakespeare.lit/desktop'
          to='hecate@shakespeare.lit'>
        <x xmlns='jabber:x:conference'
           jid='darkcave@macbeth.shakespeare.lit'
           password='cauldronburn'
           reason='Hey Hecate, this is the place for all good witches!'/>
      </message>

    Stanza Interface:

    ::

        jid      -- The JID of the groupchat room
        password -- The password used to gain entry in the room
                    (optional)
        reason   -- The reason for the invitation (optional)

    """

    name = "x"
    namespace = "jabber:x:conference"
    plugin_attrib = "groupchat_invite"
    interfaces = ("jid", "password", "reason")




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz,
# Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0231.stanza import BitsOfBinary
from slixmpp.plugins.xep_0231.bob import XEP_0231


register_plugin(XEP_0231)




# Slixmpp: The Slick XMPP Library
# Implementation of xeps for Internet of Things
# http://wiki.xmpp.org/web/Tech_pages/IoT_systems
# Copyright (C) 2013 Sustainable Innovation, Joachim.lindborg@sust.se, bjorn.westrom@consoden.se
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime

class Device(object):
    """
    Example implementation of a device control object.

    The device object may by any custom implementation to support
    specific devices, but it must implement the functions:
          has_control_field
          set_control_fields
    """

    def __init__(self, nodeId):
        self.nodeId = nodeId
        self.control_fields = {}

    def has_control_field(self, field, typename):
        """
        Returns true if the supplied field name exists
        and the type matches for control in this device.

        Arguments:
            field      -- The field name
            typename   -- The expected type
        """
        if field in self.control_fields and self.control_fields[field]["type"] == typename:
            return True
        return False

    def set_control_fields(self, fields, session, callback):
        """
        Starts a control setting procedure. Verifies the fields,
        sets the data and (if needed) and calls the callback.

        Arguments:
            fields   -- List of control fields in tuple format:
                        (name, typename, value)
            session  -- Session id, only used in the callback as identifier
            callback -- Callback function to call when control set is complete.

                    The callback function must support the following arguments:

                session     -- Session id, as supplied in the
                               request_fields call
                nodeId      -- Identifier for this device
                result      -- The current result status of the readout.
                               Valid values are:
                               "error"  - Set fields failed.
                               "ok"     - All fields were set.
                error_field -- [optional] Only applies when result == "error"
                               The field name that failed
                               (usually means it is missing)
                error_msg   -- [optional] Only applies when result == "error".
                               Error details when a request failed.
        """

        if len(fields) > 0:
            # Check availiability
            for name, typename, value in fields:
                if not self.has_control_field(name, typename):
                    self._send_control_reject(session, name, "NotFound", callback)
                    return False

        for name, typename, value in fields:
            self._set_field_value(name, value)

        callback(session, result="ok", nodeId=self.nodeId)
        return True

    def _send_control_reject(self, session, field, message, callback):
        """
        Sends a reject to the caller

        Arguments:
            session  -- Session id, see definition in
                        set_control_fields function
            callback -- Callback function, see definition in
                        set_control_fields function
        """
        callback(session, result="error", nodeId=self.nodeId, error_field=field, error_msg=message)

    def _add_control_field(self, name, typename, value):
        """
        Adds a control field to the device

        Arguments:
            name     -- Name of the field
            typename -- Type of the field, one of:
                        (boolean, color, string, date, dateTime,
                         double, duration, int, long, time)
            value    -- Field value
        """
        self.control_fields[name] = {"type": typename, "value": value}

    def _set_field_value(self, name, value):
        """
        Set the value of a control field

        Arguments:
            name     -- Name of the field
            value    -- New value for the field
        """
        if name in self.control_fields:
            self.control_fields[name]["value"] = value

    def _get_field_value(self, name):
        """
        Get the value of a control field. Only used for unit testing.

        Arguments:
            name     -- Name of the field
        """
        if name in self.control_fields:
            return self.control_fields[name]["value"]
        return None




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Encryption(ElementBase):
    name = 'encryption'
    namespace = 'urn:xmpp:eme:0'
    plugin_attrib = 'eme'
    interfaces = {'namespace', 'name'}



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza.error import Error
from slixmpp.stanza.iq import Iq
from slixmpp.stanza.message import Message
from slixmpp.stanza.presence import Presence
from slixmpp.stanza.stream_features import StreamFeatures
from slixmpp.stanza.stream_error import StreamError
from slixmpp.stanza.handshake import Handshake

__all__ = [
    'Error', 'Iq', 'Message', 'Presence', 'StreamFeatures', 'StreamError',
    'Handshake'
]



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import Iterable, Optional

from slixmpp import JID
from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Presence
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback

from slixmpp.plugins.xep_0437 import stanza


class XEP_0437(BasePlugin):
    name = 'xep_0437'
    description = 'XEP-0437: Room Activity Indicators'
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self):
        stanza.register_plugins()
        self.xmpp.register_handler(Callback(
            'RAI Received',
            StanzaPath("presence/rai"),
            self._handle_rai,
        ))
        self.xmpp.register_handler(Callback(
            'RAI Activity Received',
            StanzaPath("presence/rai/activity"),
            self._handle_rai_activity,
        ))

    def plugin_end(self):
        self.xmpp.remove_handler('RAI received')
        self.xmpp.remove_handler('RAI Activity received')

    def _handle_rai(self, presence: Presence):
        self.xmpp.event('room_activity_bare', presence)

    def _handle_rai_activity(self, presence: Presence):
        self.xmpp.event('room_activity', presence)

    def subscribe(self, service: JID, *,
                  pfrom: Optional[JID] = None):
        """
        Subscribe to room activty on a MUC service.
        :param JID service: MUC service
        """
        pres = self.xmpp.make_presence(pto=service, pfrom=pfrom)
        pres.enable('rai')
        pres.send()

    def unsubscribe(self, service: JID, *,
                    pfrom: Optional[JID] = None):
        """
        Unsubscribe from room activty on a MUC service.
        :param JID service: MUC service
        """
        pres = self.xmpp.make_presence(
            pto=service, pfrom=pfrom, ptype='unavailable',
        )
        pres.send()



from slixmpp.xmlstream import ElementBase, ET


class FormValidation(ElementBase):
    """
    Validation values for form fields.

    Example:

    ::

        <field var='evt.date' type='text-single' label='Event Date/Time'>
          <validate xmlns='http://jabber.org/protocol/xdata-validate'
                    datatype='xs:dateTime'/>
          <value>2003-10-06T11:22:00-07:00</value>
        </field>

    Questions:

    * Should this look at the datatype value and convert the range values as appropriate?
    * Should this stanza provide a pass/fail for a value from the field, or convert field value to datatype?
    """

    namespace = 'http://jabber.org/protocol/xdata-validate'
    name = 'validate'
    plugin_attrib = 'validate'
    interfaces = {'datatype', 'basic', 'open', 'range', 'regex', }
    sub_interfaces = {'basic', 'open', 'range', 'regex', }
    plugin_attrib_map = {}
    plugin_tag_map = {}

    def _add_field(self, name):
        self.remove_all()
        item_xml = ET.Element('{%s}%s' % (self.namespace, name))
        self.xml.append(item_xml)
        return item_xml

    def set_basic(self, value):
        if value:
            self._add_field('basic')
        else:
            del self['basic']

    def set_open(self, value):
        if value:
            self._add_field('open')
        else:
            del self['open']

    def set_regex(self, regex):
        if regex:
            _regex = self._add_field('regex')
            _regex.text = regex
        else:
            del self['regex']

    def set_range(self, value, minimum=None, maximum=None):
        if value:
            _range = self._add_field('range')
            _range.attrib['min'] = str(minimum)
            _range.attrib['max'] = str(maximum)
        else:
            del self['range']

    def remove_all(self, except_tag=None):
        for a in self.sub_interfaces:
            if a != except_tag:
                del self[a]

    def get_basic(self):
        present = self.xml.find('{%s}basic' % self.namespace)
        return present is not None

    def get_open(self):
        present = self.xml.find('{%s}open' % self.namespace)
        return present is not None

    def get_regex(self):
        present = self.xml.find('{%s}regex' % self.namespace)
        if present is not None:
            return present.text

        return False

    def get_range(self):
        present = self.xml.find('{%s}range' % self.namespace)
        if present is not None:
            attributes = present.attrib
            return_value = dict()
            if 'min' in attributes:
                return_value['minimum'] = attributes['min']
            if 'max' in attributes:
                return_value['maximum'] = attributes['max']
            return return_value

        return False



import time

from slixmpp import Message
import unittest
from slixmpp.test import SlixTest


class TestFilters(SlixTest):

    """
    Test using incoming and outgoing filters.
    """

    def setUp(self):
        self.stream_start()

    def testIncoming(self):

        data = []

        def in_filter(stanza):
            if isinstance(stanza, Message):
                if stanza['body'] == 'testing':
                    stanza['subject'] = stanza['body'] + ' filter'
                    print('>>> %s' % stanza['subject'])
            return stanza

        def on_message(msg):
            print('<<< %s' % msg['subject'])
            data.append(msg['subject'])

        self.xmpp.add_filter('in', in_filter)
        self.xmpp.add_event_handler('message', on_message)

        self.recv("""
          <message>
            <body>no filter</body>
          </message>
        """)

        self.recv("""
          <message>
            <body>testing</body>
          </message>
        """)

        self.assertEqual(data, ['', 'testing filter'],
                'Incoming filter did not apply %s' % data)

    def testOutgoing(self):

        def out_filter(stanza):
            if isinstance(stanza, Message):
                if stanza['body'] == 'testing':
                    stanza['body'] = 'changed!'
            return stanza

        self.xmpp.add_filter('out', out_filter)

        m1 = self.Message()
        m1['body'] = 'testing'
        m1.send()

        m2 = self.Message()
        m2['body'] = 'blah'
        m2.send()

        self.send("""
          <message>
            <body>changed!</body>
          </message>
        """)

        self.send("""
          <message>
            <body>blah</body>
          </message>
        """)



suite = unittest.TestLoader().loadTestsFromTestCase(TestFilters)



import unittest

from slixmpp import Message
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0428 import stanza

from slixmpp.plugins import xep_0461
from slixmpp.plugins import xep_0444


class TestFallback(SlixTest):
    def setUp(self):
        stanza.register_plugins()

    def testSingleFallbackBody(self):
        message = Message()
        message["fallback"]["for"] = "ns"
        message["fallback"]["body"]["start"] = 0
        message["fallback"]["body"]["end"] = 8

        self.check(
            message,  # language=XML
            """
            <message>
              <fallback xmlns='urn:xmpp:fallback:0' for='ns'>
                <body start="0" end="8" />
              </fallback>
            </message>
            """,
        )

    def testSingleFallbackSubject(self):
        message = Message()
        message["fallback"]["for"] = "ns"
        message["fallback"]["subject"]["start"] = 0
        message["fallback"]["subject"]["end"] = 8

        self.check(
            message,  # language=XML
            """
            <message>
              <fallback xmlns='urn:xmpp:fallback:0' for='ns'>
                <subject start="0" end="8" />
              </fallback>
            </message>
            """,
        )

    def testSingleFallbackWholeBody(self):
        message = Message()
        message["fallback"]["for"] = "ns"
        message["fallback"].enable("body")
        self.check(
            message,  # language=XML
            """
            <message>
              <fallback xmlns='urn:xmpp:fallback:0' for='ns'>
                <body />
              </fallback>
            </message>
            """,
        )

    def testMultiFallback(self):
        message = Message()

        f1 = stanza.Fallback()
        f1["for"] = "ns1"

        f2 = stanza.Fallback()
        f2["for"] = "ns2"

        message.append(f1)
        message.append(f2)

        self.check(
            message,  # language=XML
            """
            <message>
              <fallback xmlns='urn:xmpp:fallback:0' for='ns1' />
              <fallback xmlns='urn:xmpp:fallback:0' for='ns2' />
            </message>
            """,
        )

        for i, fallback in enumerate(message["fallbacks"], start=1):
            self.assertEqual(fallback["for"], f"ns{i}")

    def testStripFallbackPartOfBody(self):
        message = Message()
        message["body"] = "> quoted\nsome-body"
        message["fallback"]["for"] = xep_0461.stanza.NS
        message["fallback"]["body"]["start"] = 0
        message["fallback"]["body"]["end"] = 9

        self.check(
            message,  # language=XML
            """
            <message>
              <body>&gt; quoted\nsome-body</body>
              <fallback xmlns='urn:xmpp:fallback:0' for='urn:xmpp:reply:0'>
                <body start="0" end="9" />
              </fallback>
            </message>
            """,
        )

        self.assertEqual(
            message["fallback"].get_stripped_body(xep_0461.stanza.NS), "some-body"
        )

    def testStripWholeBody(self):
        message = Message()
        message["body"] = "> quoted\nsome-body"
        message["fallback"]["for"] = "ns"
        message["fallback"].enable("body")

        self.check(
            message,  # language=XML
            """
            <message>
              <body>&gt; quoted\nsome-body</body>
              <fallback xmlns='urn:xmpp:fallback:0' for='ns'>
                <body />
              </fallback>
            </message>
            """,
        )

        self.assertEqual(message["fallback"].get_stripped_body("ns"), "")

    def testStripMultiFallback(self):
        message = Message()
        message["body"] = "> huuuuu\n👍"

        message["fallback"]["for"] = xep_0461.stanza.NS
        message["fallback"]["body"]["start"] = 0
        message["fallback"]["body"]["end"] = 9

        reaction_fallback = stanza.Fallback()
        reaction_fallback["for"] = xep_0444.stanza.NS
        reaction_fallback.enable("body")
        message.append(reaction_fallback)

        self.assertEqual(message["fallback"].get_stripped_body(xep_0461.stanza.NS), "👍")
        self.assertEqual(message["fallback"].get_stripped_body(xep_0444.stanza.NS), "")


suite = unittest.TestLoader().loadTestsFromTestCase(TestFallback)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ET


class OptionalSetting(object):

    interfaces = {'required'}

    def set_required(self, value):
        if value in (True, 'true', 'True', '1'):
            self.xml.append(ET.Element("{%s}required" % self.namespace))
        elif self['required']:
            self.del_required()

    def get_required(self):
        required = self.xml.find("{%s}required" % self.namespace)
        return required is not None

    def del_required(self):
        required = self.xml.find("{%s}required" % self.namespace)
        if required is not None:
            self.xml.remove(required)



import unittest
import logging

from slixmpp.plugins.base import PluginManager, BasePlugin, register_plugin


class A(BasePlugin):
    name = 'a'


class B(BasePlugin):
    name = 'b'


class C(BasePlugin):
    name = 'c'
    dependencies = {'b', 'd'}


class D(BasePlugin):
    name = 'd'
    dependencies = {'c'}


class E(BasePlugin):
    name = 'e'
    dependencies = {'a', 'd'}

class F(BasePlugin):
    name = 'f'
    dependencies = {'a', 'b'}


register_plugin(A)
register_plugin(B)
register_plugin(C)
register_plugin(D)
register_plugin(E)
register_plugin(F)


class TestPlugins(unittest.TestCase):


    def test_enable(self):
        """Enable a single plugin."""
        p = PluginManager(None)

        events = []

        def init(self):
            events.append('init')

        A.plugin_init = init

        p.enable('a')

        self.assertEqual(len(p), 1, "Wrong number of enabled plugins.")
        self.assertEqual(events, ['init'], "Plugin init method not called.")

    def test_disable(self):
        """Disable a single plugin."""
        p = PluginManager(None)

        events = []

        def init(self):
            events.append('init')

        def end(self):
            events.append('end')

        A.plugin_init = init
        A.plugin_end = end

        p.enable('a')
        p.disable('a')

        self.assertEqual(len(p), 0, "Wrong number of enabled plugins.")
        self.assertEqual(events, ['init', 'end'],
                "Plugin lifecycle methods not called.")

    def test_enable_dependencies(self):
        """Enable a plugin with acyclic dependencies."""
        p = PluginManager(None)

        events = []

        A.plugin_init = lambda s: events.append('init_a')
        B.plugin_init = lambda s: events.append('init_b')

        p.enable('f')

        self.assertEqual(len(p), 3, "Wrong number of enabled plugins.")
        self.assertTrue('init_a' in events, "Dependency A not enabled.")
        self.assertTrue('init_b' in events, "Dependency B not enabled.")

    def test_enable_cyclic_dependencies(self):
        """Enable a plugin with cyclic dependencies."""

        p = PluginManager(None)

        events = []

        B.plugin_init = lambda s: events.append('init_b')
        C.plugin_init = lambda s: events.append('init_c')
        D.plugin_init = lambda s: events.append('init_d')

        p.enable('c')

        self.assertEqual(len(p), 3, "Wrong number of enabled plugins.")
        self.assertTrue('init_b' in events, "Dependency B not enabled.")
        self.assertTrue('init_c' in events, "Dependency C not enabled.")
        self.assertTrue('init_d' in events, "Dependency D not enabled.")

    def test_disable_dependendents(self):
        """Disable a plugin with dependents."""

        p = PluginManager(None)

        events = []

        A.plugin_end = lambda s: events.append('end_a')
        B.plugin_end = lambda s: events.append('end_b')
        F.plugin_end = lambda s: events.append('end_f')

        p.enable('f')
        p.disable('a')

        self.assertEqual(len(p), 1, "Wrong number of enabled plugins.")
        self.assertTrue('end_f' in events, "Dependent F not disabled.")
        self.assertTrue('end_a' in events, "Plugin A not disabled.")

    def test_disable_cyclic_dependents(self):
        """Disable a plugin with cyclic dependents."""

        p = PluginManager(None)

        events = []

        B.plugin_end = lambda s: events.append('end_b')
        C.plugin_end = lambda s: events.append('end_c')
        D.plugin_end = lambda s: events.append('end_d')

        p.enable('c')
        p.disable('b')

        self.assertEqual(len(p), 0, "Wrong number of enabled plugins.")
        self.assertTrue('end_b' in events, "Plugin B not disabled.")
        self.assertTrue('end_c' in events, "Dependent C not disabled.")
        self.assertTrue('end_d' in events, "Dependent D not disabled.")



suite = unittest.TestLoader().loadTestsFromTestCase(TestPlugins)

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG,
                        format='%(levelname)-8s %(message)s')

    tests = unittest.TestSuite([suite])
    unittest.TextTestRunner(verbosity=2).run(tests)



import unittest
from slixmpp.jid import JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0410 import PingStatus


class TestStreamSelfPing(SlixTest):

    def test_ping_workflow(self):
        """
        Try the autoping provided by the plugin, as well as the events
        generated
        """
        self.stream_start(mode='client', plugins=['xep_0410'])
        self.xmpp.plugin['xep_0410'].ping_interval = 0.1

        recv = []
        def recv_ping(args):
            recv.append(args)

        self.xmpp.add_event_handler('muc_ping_changed', recv_ping)

        self.xmpp.plugin['xep_0410'].enable_self_ping(
            JID('muc@muc.example.com/toto'),
            JID('toto@example.com/test'),
        )
        self.wait_(0.2)
        self.send("""
<iq from="toto@example.com/test" to="muc@muc.example.com/toto" type="get" id="1">
    <ping xmlns='urn:xmpp:ping'/>
</iq>
        """)
        self.recv("""
<iq to="toto@example.com/test" from="muc@muc.example.com/toto" type="result" id="1"/>
        """)
        self.wait_()

        self.assertEqual(
            recv[-1]['result'],
            PingStatus.JOINED,
        )
        self.assertEqual(
            self.xmpp.plugin['xep_0410'].get_ping_status(
                JID('muc@muc.example.com/toto'),
                JID('toto@example.com/test'),
            ),
            PingStatus.JOINED,
        )
        self.wait_(0.2)
        self.send("""
<iq from="toto@example.com/test" to="muc@muc.example.com/toto" type="get" id="2">
    <ping xmlns='urn:xmpp:ping'/>
</iq>
        """)
        self.recv("""
<iq to="toto@example.com/test" from="muc@muc.example.com/toto" type="error" id="2">
  <error type="cancel" by="muc.example.com">
    <not-acceptable xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
  </error>
</iq>
        """)
        self.wait_()
        self.assertEqual(
            recv[-1]['result'],
            PingStatus.DISCONNECTED,
        )
        self.assertEqual(
            self.xmpp.plugin['xep_0410'].get_ping_status(
                JID('muc@muc.example.com/toto'),
                JID('toto@example.com/test'),
            ),
            PingStatus.DISCONNECTED,
        )
        self.xmpp.plugin['xep_0410'].disable_self_ping(
            JID('muc@muc.example.com/toto'),
            JID('toto@example.com/test'),
        )
        self.wait_()



suite = unittest.TestLoader().loadTestsFromTestCase(TestStreamSelfPing)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import copy
import logging

from slixmpp.thirdparty import OrderedSet

from slixmpp.xmlstream import ElementBase, ET
from slixmpp.plugins.xep_0004.stanza import FormField


log = logging.getLogger(__name__)


class Form(ElementBase):
    namespace = 'jabber:x:data'
    name = 'x'
    plugin_attrib = 'form'
    plugin_multi_attrib = 'forms'
    interfaces = OrderedSet(('instructions', 'reported', 'title', 'type', 'items', 'values'))
    sub_interfaces = {'title'}
    form_types = {'cancel', 'form', 'result', 'submit'}

    def __init__(self, *args, **kwargs):
        title = None
        if 'title' in kwargs:
            title = kwargs['title']
            del kwargs['title']
        ElementBase.__init__(self, *args, **kwargs)
        if title is not None:
            self['title'] = title

    def setup(self, xml=None):
        if ElementBase.setup(self, xml):
            # If we had to generate xml
            self['type'] = 'form'

    @property
    def field(self):
        return self.get_fields()

    def set_type(self, ftype):
        self._set_attr('type', ftype)
        if ftype == 'submit':
            fields = self.get_fields()
            for var in fields:
                field = fields[var]
                if field['type'] != 'hidden':
                    del field['type']
                del field['label']
                del field['desc']
                del field['required']
                del field['options']
        elif ftype == 'cancel':
            del self['fields']

    def add_field(self, var='', ftype=None, label='', desc='',
                  required=False, value=None, options=None, **kwargs):
        kwtype = kwargs.get('type', None)
        if kwtype is None:
            kwtype = ftype

        field = FormField()
        field['var'] = var
        field['type'] = kwtype
        field['value'] = value
        if self['type'] in ('form', 'result'):
            field['label'] = label
            field['desc'] = desc
            field['required'] = required
            if options is not None:
                for option in options:
                    field.add_option(**option)
        else:
            if field['type'] != 'hidden':
                del field['type']
        self.append(field)
        return field

    def add_item(self, values):
        itemXML = ET.Element('{%s}item' % self.namespace)
        self.xml.append(itemXML)
        reported_vars = self['reported'].keys()
        for var in reported_vars:
            field = FormField()
            field._type = self['reported'][var]['type']
            field['var'] = var
            field['value'] = values.get(var, None)
            itemXML.append(field.xml)

    def add_reported(self, var, ftype=None, label='', desc='', **kwargs):
        kwtype = kwargs.get('type', None)
        if kwtype is None:
            kwtype = ftype
        reported = self.xml.find('{%s}reported' % self.namespace)
        if reported is None:
            reported = ET.Element('{%s}reported' % self.namespace)
            self.xml.append(reported)
        fieldXML = ET.Element('{%s}field' % FormField.namespace)
        reported.append(fieldXML)
        field = FormField(xml=fieldXML)
        field['var'] = var
        field['type'] = kwtype
        field['label'] = label
        field['desc'] = desc
        return field

    def cancel(self):
        self['type'] = 'cancel'

    def del_fields(self):
        fieldsXML = self.xml.findall('{%s}field' % FormField.namespace)
        for fieldXML in fieldsXML:
            self.xml.remove(fieldXML)

    def del_instructions(self):
        instsXML = self.xml.findall('{%s}instructions')
        for instXML in instsXML:
            self.xml.remove(instXML)

    def del_items(self):
        itemsXML = self.xml.find('{%s}item' % self.namespace)
        for itemXML in itemsXML:
            self.xml.remove(itemXML)

    def del_reported(self):
        reportedXML = self.xml.find('{%s}reported' % self.namespace)
        if reportedXML is not None:
            self.xml.remove(reportedXML)

    def get_fields(self, use_dict=False):
        fields = {}
        for stanza in self['substanzas']:
            if isinstance(stanza, FormField):
                fields[stanza['var']] = stanza
        return fields

    def get_instructions(self):
        instsXML = self.xml.findall('{%s}instructions' % self.namespace)
        return "\n".join([instXML.text for instXML in instsXML])

    def get_items(self):
        items = []
        itemsXML = self.xml.findall('{%s}item' % self.namespace)
        for itemXML in itemsXML:
            item = {}
            fieldsXML = itemXML.findall('{%s}field' % FormField.namespace)
            for fieldXML in fieldsXML:
                field = FormField(xml=fieldXML)
                item[field['var']] = field['value']
            items.append(item)
        return items

    def get_reported(self):
        fields = {}
        xml = self.xml.findall('{%s}reported/{%s}field' % (self.namespace,
                                                           FormField.namespace))
        for field in xml:
            field = FormField(xml=field)
            fields[field['var']] = field
        return fields

    def get_values(self):
        values = {}
        fields = self.get_fields()
        for var in fields:
            values[var] = fields[var]['value']
        return values

    def reply(self):
        if self['type'] == 'form':
            self['type'] = 'submit'
        elif self['type'] == 'submit':
            self['type'] = 'result'

    def set_fields(self, fields):
        del self['fields']
        if not isinstance(fields, list):
            fields = fields.items()
        for var, field in fields:
            field['var'] = var
            self.add_field(
                var=field.get('var'),
                label=field.get('label'),
                desc=field.get('desc'),
                required=field.get('required'),
                value=field.get('value'),
                options=field.get('options'),
                type=field.get('type'))

    def set_instructions(self, instructions):
        del self['instructions']
        if instructions in [None, '']:
            return
        if not isinstance(instructions, list):
            instructions = instructions.split('\n')
        for instruction in instructions:
            inst = ET.Element('{%s}instructions' % self.namespace)
            inst.text = instruction
            self.xml.append(inst)

    def set_items(self, items):
        for item in items:
            self.add_item(item)

    def set_reported(self, reported):
        """
        This either needs a dictionary of dictionaries or a dictionary of form fields.
        :param reported:
        :return:
        """
        for var in reported:
            field = reported[var]

            if isinstance(field, dict):
                self.add_reported(**field)
            else:
                reported = self.xml.find('{%s}reported' % self.namespace)
                if reported is None:
                    reported = ET.Element('{%s}reported' % self.namespace)
                    self.xml.append(reported)

                fieldXML = ET.Element('{%s}field' % FormField.namespace)
                reported.append(fieldXML)
                new_field = FormField(xml=fieldXML)
                new_field.values = field.values

    def set_values(self, values):
        fields = self.get_fields()
        for field in values:
            if field not in self.get_fields():
                fields[field] = self.add_field(var=field)
            self.get_fields()[field]['value'] = values[field]

    def merge(self, other):
        new = copy.copy(self)
        if type(other) == dict:
            new['values'] = other
            return new
        nfields = new['fields']
        ofields = other['fields']
        nfields.update(ofields)
        new['fields'] = nfields
        return new


Form.addField = Form.add_field
Form.addReported = Form.add_reported
Form.delFields = Form.del_fields
Form.delInstructions = Form.del_instructions
Form.delReported = Form.del_reported
Form.getFields = Form.get_fields
Form.getInstructions = Form.get_instructions
Form.getReported = Form.get_reported
Form.getValues = Form.get_values
Form.setFields = Form.set_fields
Form.setInstructions = Form.set_instructions
Form.setReported = Form.set_reported
Form.setValues = Form.set_values



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import StanzaBase
from typing import Optional


class Handshake(StanzaBase):

    """
    Jabber Component protocol handshake
    """
    namespace = 'jabber:component:accept'
    name = 'handshake'
    interfaces = {'value'}

    def set_value(self, value: str):
        self.xml.text = value

    def get_value(self) -> Optional[str]:
        return self.xml.text

    def del_value(self):
        self.xml.text = ''




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0079.stanza import (
        AMP, Rule, InvalidRules, UnsupportedConditions,
        UnsupportedActions, FailedRules, FailedRule,
        AMPFeature)
from slixmpp.plugins.xep_0079.amp import XEP_0079


register_plugin(XEP_0079)



import unittest
from slixmpp.test.integration import SlixIntegration
from slixmpp.plugins.xep_0439 import stanza


class TestQuickResponse(SlixIntegration):
    async def asyncSetUp(self):
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0439'])
        await self.connect_clients()

    async def test_quickresponse(self):
        """Send and receive actions and responses"""
        actions = [
            ('id1', 'Action 1'),
            ('id2', 'Action 2'),
        ]
        self.clients[0]['xep_0439'].ask_for_actions(
            self.clients[1].boundjid.full,
            "Action 1 or 2 ?",
            actions
        )
        msg = await self.clients[1].wait_until('action_received')
        actions_recv = [
            (st['id'], st['label']) for st in msg if isinstance(st, stanza.Action)
        ]
        self.assertEqual(
                actions,
                actions_recv,
        )

        reply = self.clients[1].make_message(
            mto=self.clients[0].boundjid.full
        )
        reply['action_selected']['id'] = 'id1'
        reply.send()

        reply_recv = await self.clients[0].wait_until('action_selected')
        self.assertEqual(
            reply_recv['action_selected']['id'],
            'id1',
        )

        self.clients[0]['xep_0439'].ask_for_response(
            self.clients[1].boundjid.full,
            "Reply with action 1 or 2 (id1/id2) ?",
            actions
        )
        msg = await self.clients[1].wait_until('responses_received')


suite = unittest.TestLoader().loadTestsFromTestCase(TestQuickResponse)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from datetime import datetime, timezone

from typing import Optional

from slixmpp import JID
from slixmpp.stanza import Presence
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0319 import stanza


def get_local_timezone():
    return datetime.now(timezone.utc).astimezone().tzinfo


class XEP_0319(BasePlugin):
    name = 'xep_0319'
    description = 'XEP-0319: Last User Interaction in Presence'
    dependencies = {'xep_0012'}
    stanza = stanza

    def plugin_init(self):
        self._idle_stamps = {}
        register_stanza_plugin(Presence, stanza.Idle)
        self.api.register(self._set_idle, 'set_idle', default=True)
        self.api.register(self._get_idle, 'get_idle', default=True)
        self.xmpp.register_handler(Callback(
            'Idle Presence',
            StanzaPath('presence/idle'),
            self._idle_presence
        ))
        self.xmpp.add_filter('out', self._stamp_idle_presence)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:idle:1')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:idle:1')
        self.xmpp.del_filter('out', self._stamp_idle_presence)
        self.xmpp.remove_handler('Idle Presence')

    async def idle(self, jid: Optional[JID] = None,
                   since: Optional[datetime] = None):
        """Set an idle duration for a JID

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        seconds = None
        timezone = get_local_timezone()
        if since is None:
            since = datetime.now(timezone)
        else:
            seconds = datetime.now(timezone) - since
        await self.api['set_idle'](jid, None, None, since)
        await self.xmpp['xep_0012'].set_last_activity(jid=jid, seconds=seconds)

    async def active(self, jid: Optional[JID] = None):
        """Reset the idle timer.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        await self.api['set_idle'](jid, None, None, None)
        await self.xmpp['xep_0012'].del_last_activity(jid)

    def _set_idle(self, jid, node, ifrom, data):
        self._idle_stamps[jid] = data

    def _get_idle(self, jid, node, ifrom, data):
        return self._idle_stamps.get(jid, None)

    def _idle_presence(self, pres):
        self.xmpp.event('presence_idle', pres)

    async def _stamp_idle_presence(self, stanza):
        if isinstance(stanza, Presence):
            since = await self.api['get_idle'](stanza['from'] or self.xmpp.boundjid)
            if since:
                stanza['idle']['since'] = since
        return stanza




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ET, ElementBase


class UserSettings(ElementBase):
    name = 'usersetting'
    namespace = 'google:setting'
    plugin_attrib = 'google_settings'
    interfaces = set(['auto_accept_suggestions',
                      'mail_notifications',
                      'archiving_enabled',
                      'gmail',
                      'email_verified',
                      'domain_privacy_notice',
                      'display_name'])

    def _get_setting(self, setting):
        xml = self.xml.find('{%s}%s' % (self.namespace, setting))
        if xml is not None:
            return xml.attrib.get('value', '') == 'true'
        return False

    def _set_setting(self, setting, value):
        self._del_setting(setting)
        if value in (True, False):
            xml = ET.Element('{%s}%s' % (self.namespace, setting))
            xml.attrib['value'] = 'true' if value else 'false'
            self.xml.append(xml)

    def _del_setting(self, setting):
        xml = self.xml.find('{%s}%s' % (self.namespace, setting))
        if xml is not None:
            self.xml.remove(xml)

    def get_display_name(self):
        xml = self.xml.find('{%s}%s' % (self.namespace, 'displayname'))
        if xml is not None:
            return xml.attrib.get('value', '')
        return ''

    def set_display_name(self, value):
        self._del_setting(setting)
        if value:
            xml = ET.Element('{%s}%s' % (self.namespace, 'displayname'))
            xml.attrib['value'] = value
            self.xml.append(xml)

    def del_display_name(self):
        self._del_setting('displayname')

    def get_auto_accept_suggestions(self):
        return self._get_setting('autoacceptsuggestions')

    def get_mail_notifications(self):
        return self._get_setting('mailnotifications')

    def get_archiving_enabled(self):
        return self._get_setting('archivingenabled')

    def get_gmail(self):
        return self._get_setting('gmail')

    def get_email_verified(self):
        return self._get_setting('emailverified')

    def get_domain_privacy_notice(self):
        return self._get_setting('domainprivacynotice')

    def set_auto_accept_suggestions(self, value):
        self._set_setting('autoacceptsuggestions', value)

    def set_mail_notifications(self, value):
        self._set_setting('mailnotifications', value)

    def set_archiving_enabled(self, value):
        self._set_setting('archivingenabled', value)

    def set_gmail(self, value):
        self._set_setting('gmail', value)

    def set_email_verified(self, value):
        self._set_setting('emailverified', value)

    def set_domain_privacy_notice(self, value):
        self._set_setting('domainprivacynotice', value)

    def del_auto_accept_suggestions(self):
        self._del_setting('autoacceptsuggestions')

    def del_mail_notifications(self):
        self._del_setting('mailnotifications')

    def del_archiving_enabled(self):
        self._del_setting('archivingenabled')

    def del_gmail(self):
        self._del_setting('gmail')

    def del_email_verified(self):
        self._del_setting('emailverified')

    def del_domain_privacy_notice(self):
        self._del_setting('domainprivacynotice')



import unittest
from slixmpp.test.integration import SlixIntegration


class TestReactions(SlixIntegration):
    async def asyncSetUp(self):
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0444'])
        await self.connect_clients()

    async def test_send_reaction(self):
        """Make sure we can send and receive reactions"""
        self.clients[0]['xep_0444'].send_reactions(
            self.clients[1].boundjid.full,
            to_id='toto',
            reactions=['🦙', '🦦'],
        )
        msg = await self.clients[1].wait_until('reactions')
        self.assertEqual(
            msg['reactions'].get_values(),
            {'🦙', '🦦'},
        )
        self.assertEqual(msg['reactions']['id'], 'toto')

suite = unittest.TestLoader().loadTestsFromTestCase(TestReactions)



import unittest

from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0446 import stanza


class TestFileMeta(SlixTest):
    def setUp(self):
        stanza.register_plugins()

    def test_simple(self):
        file = stanza.File()
        file["desc"] = "a description"
        file["name"] = "toto.jpg"
        file["media-type"] = "image/jpeg"
        file["height"] = 1024
        file["width"] = 768
        file["size"] = 2048
        self.check(
            file,
            """
            <file xmlns='urn:xmpp:file:metadata:0'>
              <desc>a description</desc>
              <name>toto.jpg</name>
              <media-type>image/jpeg</media-type>
              <height>1024</height>
              <width>768</width>
              <size>2048</size>
            </file>
            """,
        )

    def test_bad_value(self):
        file = stanza.File()
        file["desc"] = "My great video"
        file["name"] = "toto.mp4"
        file["media-type"] = "video/3gpp"
        file["height"] = 1024
        file["width"] = 768
        with self.assertRaises(ValueError):
            file["length"] = -100

    def test_hash_element(self):
        file = stanza.File()
        file["desc"] = "My great video"
        file["name"] = "toto.3gp"
        file["media-type"] = "video/3gpp"
        file["height"] = 1024
        file["width"] = 768
        file["length"] = 2000
        file["hash"]["algo"] = "sha3-256"
        file["hash"]["value"] = "abcdef="
        self.check(
            file,
            """
            <file xmlns='urn:xmpp:file:metadata:0'>
              <desc>My great video</desc>
              <name>toto.3gp</name>
              <media-type>video/3gpp</media-type>
              <height>1024</height>
              <width>768</width>
              <length>2000</length>
              <hash xmlns='urn:xmpp:hashes:2' algo="sha3-256">abcdef=</hash>
            </file>
            """,
        )

    def test_thumbnail_element(self):
        file = stanza.File()
        file["desc"] = "a description"
        file["name"] = "toto.jpg"
        file["media-type"] = "image/jpeg"
        file["height"] = 1024
        file["width"] = 768
        file["size"] = 2048
        file["thumbnail"]["media-type"] = "image/png"
        file["thumbnail"]["uri"] = "cid:sha1+deadbeef@bob.xmpp.org"
        file["thumbnail"]["width"] = 128
        file["thumbnail"]["height"] = 96
        self.check(
            file,
            """
            <file xmlns='urn:xmpp:file:metadata:0'>
              <desc>a description</desc>
              <name>toto.jpg</name>
              <media-type>image/jpeg</media-type>
              <height>1024</height>
              <width>768</width>
              <size>2048</size>
              <thumbnail xmlns='urn:xmpp:thumbs:1'
                 uri='cid:sha1+deadbeef@bob.xmpp.org'
                 media-type='image/png'
                 width='128'
                 height='96'/>
            </file>
            """,
        )



suite = unittest.TestLoader().loadTestsFromTestCase(TestFileMeta)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp
from slixmpp.exceptions import XMPPError
from slixmpp import asyncio


FILE_TYPES = {
    'image/png': 'png',
    'image/gif': 'gif',
    'image/jpeg': 'jpg'
}


class AvatarDownloader(slixmpp.ClientXMPP):

    """
    A basic script for downloading the avatars for a user's contacts.
    """

    def __init__(self, jid, password):
        slixmpp.ClientXMPP.__init__(self, jid, password)
        self.add_event_handler("session_start", self.start)
        self.add_event_handler("changed_status", self.wait_for_presences)

        self.add_event_handler('vcard_avatar_update', self.on_vcard_avatar)
        self.add_event_handler('avatar_metadata_publish', self.on_avatar)

        self.received = set()
        self.presences_received = asyncio.Event()
        self.roster_received = asyncio.Event()

    def roster_received_cb(self, event):
        self.roster_received.set()
        self.presences_received.clear()

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        self.get_roster(callback=self.roster_received_cb)

        print('Waiting for presence updates...\n')
        await self.roster_received.wait()
        print('Roster received')
        await self.presences_received.wait()
        self.disconnect()

    async def on_vcard_avatar(self, pres):
        print("Received vCard avatar update from %s" % pres['from'].bare)
        try:
            result = await self['xep_0054'].get_vcard(pres['from'].bare, cached=True,
                                                           timeout=5)
        except XMPPError:
            print("Error retrieving avatar for %s" % pres['from'])
            return
        avatar = result['vcard_temp']['PHOTO']

        filetype = FILE_TYPES.get(avatar['TYPE'], 'png')
        filename = 'vcard_avatar_%s_%s.%s' % (
                pres['from'].bare,
                pres['vcard_temp_update']['photo'],
                filetype)
        with open(filename, 'wb+') as img:
            img.write(avatar['BINVAL'])

    async def on_avatar(self, msg):
        print("Received avatar update from %s" % msg['from'])
        metadata = msg['pubsub_event']['items']['item']['avatar_metadata']
        for info in metadata['items']:
            if not info['url']:
                try:
                    result = await self['xep_0084'].retrieve_avatar(msg['from'].bare, info['id'],
                                                                         timeout=5)
                except XMPPError:
                    print("Error retrieving avatar for %s" % msg['from'])
                    return

                avatar = result['pubsub']['items']['item']['avatar_data']

                filetype = FILE_TYPES.get(metadata['type'], 'png')
                filename = 'avatar_%s_%s.%s' % (msg['from'].bare, info['id'], filetype)
                with open(filename, 'wb+') as img:
                    img.write(avatar['value'])
            else:
                # We could retrieve the avatar via HTTP, etc here instead.
                pass

    def wait_for_presences(self, pres):
        """
        Wait to receive updates from all roster contacts.
        """
        self.received.add(pres['from'].bare)
        print((len(self.received), len(self.client_roster.keys())))
        if len(self.received) >= len(self.client_roster.keys()):
            self.presences_received.set()
        else:
            self.presences_received.clear()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()
    parser.add_argument("-q","--quiet", help="set logging to ERROR",
                        action="store_const",
                        dest="loglevel",
                        const=logging.ERROR,
                        default=logging.ERROR)
    parser.add_argument("-d","--debug", help="set logging to DEBUG",
                        action="store_const",
                        dest="loglevel",
                        const=logging.DEBUG,
                        default=logging.ERROR)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    xmpp = AvatarDownloader(args.jid, args.password)
    xmpp.register_plugin('xep_0054')
    xmpp.register_plugin('xep_0153')
    xmpp.register_plugin('xep_0084')

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_until_complete(xmpp.disconnected)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import hashlib
import logging
from asyncio import Future
from typing import (
    Dict,
    Optional,
)

from slixmpp import JID
from slixmpp.stanza import Presence
from slixmpp.exceptions import XMPPError, IqTimeout, IqError
from slixmpp.xmlstream import register_stanza_plugin, ElementBase
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0153 import stanza, VCardTempUpdate


log = logging.getLogger(__name__)


class XEP_0153(BasePlugin):

    name = 'xep_0153'
    description = 'XEP-0153: vCard-Based Avatars'
    dependencies = {'xep_0054'}
    stanza = stanza

    def plugin_init(self):
        self._hashes = {}

        register_stanza_plugin(Presence, VCardTempUpdate)

        self.xmpp.add_filter('out', self._update_presence)

        self.xmpp.add_event_handler('session_start', self._start)

        self.xmpp.add_event_handler('presence_available', self._recv_presence)
        self.xmpp.add_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.add_event_handler('presence_xa', self._recv_presence)
        self.xmpp.add_event_handler('presence_chat', self._recv_presence)
        self.xmpp.add_event_handler('presence_away', self._recv_presence)

        self.api.register(self._set_hash, 'set_hash', default=True)
        self.api.register(self._get_hash, 'get_hash', default=True)
        self.api.register(self._reset_hash, 'reset_hash', default=True)

    def plugin_end(self):
        self.xmpp.del_filter('out', self._update_presence)
        self.xmpp.del_event_handler('session_start', self._start)
        self.xmpp.del_event_handler('session_end', self._end)
        self.xmpp.del_event_handler('presence_available', self._recv_presence)
        self.xmpp.del_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.del_event_handler('presence_xa', self._recv_presence)
        self.xmpp.del_event_handler('presence_chat', self._recv_presence)
        self.xmpp.del_event_handler('presence_away', self._recv_presence)

    def set_avatar(self, jid: Optional[JID] = None,
                   avatar: Optional[bytes] = None,
                   mtype: Optional[str] = None, **iqkwargs) -> Future:
        """Set a VCard avatar.

        :param jid: The JID to set the avatar for.
        :param avatar: Avatar content.
        :param mtype: Avatar file type (e.g. image/jpeg).
        """
        if jid is None:
            jid = self.xmpp.boundjid.bare
        async def get_and_set_avatar():
            timeout = iqkwargs.get('timeout', None)
            timeout_cb = iqkwargs.get('timeout_callback', None)
            try:
                result = await self.xmpp['xep_0054'].get_vcard(
                    jid,
                    cached=False,
                    timeout=timeout
                )
            except IqTimeout as exc:
                if timeout_cb is not None:
                    timeout_cb(exc)
                raise
            vcard = result['vcard_temp']
            vcard['PHOTO']['TYPE'] = mtype
            vcard['PHOTO']['BINVAL'] = avatar

            try:
                result = await self.xmpp['xep_0054'].publish_vcard(
                    jid=jid,
                    vcard=vcard,
                    **iqkwargs
                )
            except IqTimeout as exc:
                timeout_cb(exc)
                raise
            await self.api['reset_hash'](jid)
            self.xmpp.roster[jid].send_last_presence()

        return self.xmpp.wrap(get_and_set_avatar())

    async def _start(self, event):
        try:
            vcard = await self.xmpp['xep_0054'].get_vcard(self.xmpp.boundjid.bare)
            data = vcard['vcard_temp']['PHOTO']['BINVAL']
            if not data:
                new_hash = ''
            else:
                new_hash = hashlib.sha1(data).hexdigest()
            await self.api['set_hash'](self.xmpp.boundjid, args=new_hash)
        except XMPPError:
            log.debug('Could not retrieve vCard for %s', self.xmpp.boundjid.bare)

    async def _update_presence(self, stanza: ElementBase) -> ElementBase:
        if not isinstance(stanza, Presence):
            return stanza

        if stanza['type'] not in ('available', 'dnd', 'chat', 'away', 'xa'):
            return stanza

        current_hash = await self.api['get_hash'](stanza['from'])
        stanza['vcard_temp_update']['photo'] = current_hash
        return stanza

    async def _recv_presence(self, pres: Presence):
        try:
            if pres.get_plugin('muc', check=True):
                # Don't process vCard avatars for MUC occupants
                # since they all share the same bare JID.
                return
        except:
            pass

        if not pres.match('presence/vcard_temp_update'):
            await self.api['set_hash'](pres['from'], args=None)
            return

        data = pres['vcard_temp_update']['photo']
        if data is None:
            return
        self.xmpp.event('vcard_avatar_update', pres)

    # =================================================================

    async def _reset_hash(self, jid: JID, node: str, ifrom: JID, args: Dict):
        own_jid = (jid.bare == self.xmpp.boundjid.bare)
        if self.xmpp.is_component:
            own_jid = (jid.domain == self.xmpp.boundjid.domain)

        await self.api['set_hash'](jid, args=None)
        if own_jid:
            self.xmpp.roster[jid].send_last_presence()

        try:
            iq = await self.xmpp['xep_0054'].get_vcard(jid=jid.bare, ifrom=ifrom)
        except (IqError, IqTimeout):
            log.debug('Could not retrieve vCard for %s', jid)
            return
        try:
            data = iq['vcard_temp']['PHOTO']['BINVAL']
        except ValueError:
            log.debug('Invalid BINVAL in vCard’s PHOTO for %s:', jid, exc_info=True)
            data = None
        if not data:
            new_hash = ''
        else:
            new_hash = hashlib.sha1(data).hexdigest()

        await self.api['set_hash'](jid, args=new_hash)

    def _get_hash(self, jid: JID, node: str, ifrom: JID, args: Dict):
        return self._hashes.get(jid.bare, None)

    def _set_hash(self, jid: JID, node: str, ifrom: JID, args: Dict):
        self._hashes[jid.bare] = args




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ET
import base64
import logging
import time

log = logging.getLogger(__name__)

_namespace = 'jabber:iq:rpc'

def fault2xml(fault):
    value = dict()
    value['faultCode'] = fault['code']
    value['faultString'] = fault['string']
    fault = ET.Element("fault", {'xmlns': _namespace})
    fault.append(_py2xml((value)))
    return fault

def xml2fault(params):
    vals = []
    for value in params.findall('{%s}value' % _namespace):
        vals.append(_xml2py(value))
    fault = dict()
    fault['code'] = vals[0]['faultCode']
    fault['string'] = vals[0]['faultString']
    return fault

def py2xml(*args):
    params = ET.Element("{%s}params" % _namespace)
    for x in args:
        param = ET.Element("{%s}param" % _namespace)
        param.append(_py2xml(x))
        params.append(param) #<params><param>...
    return params

def _py2xml(*args):
    for x in args:
        val = ET.Element("{%s}value" % _namespace)
        if x is None:
            nil = ET.Element("{%s}nil" % _namespace)
            val.append(nil)
        elif type(x) is int:
            i4 = ET.Element("{%s}i4" % _namespace)
            i4.text = str(x)
            val.append(i4)
        elif type(x) is bool:
            boolean = ET.Element("{%s}boolean" % _namespace)
            boolean.text = str(int(x))
            val.append(boolean)
        elif type(x) is str:
            string = ET.Element("{%s}string" % _namespace)
            string.text = x
            val.append(string)
        elif type(x) is float:
            double = ET.Element("{%s}double" % _namespace)
            double.text = str(x)
            val.append(double)
        elif type(x) is rpcbase64:
            b64 = ET.Element("{%s}base64" % _namespace)
            b64.text = x.encoded()
            val.append(b64)
        elif type(x) is rpctime:
            iso = ET.Element("{%s}dateTime.iso8601" % _namespace)
            iso.text = str(x)
            val.append(iso)
        elif type(x) in (list, tuple):
            array = ET.Element("{%s}array" % _namespace)
            data = ET.Element("{%s}data" % _namespace)
            for y in x:
                data.append(_py2xml(y))
            array.append(data)
            val.append(array)
        elif type(x) is dict:
            struct = ET.Element("{%s}struct" % _namespace)
            for y in x.keys():
                member = ET.Element("{%s}member" % _namespace)
                name = ET.Element("{%s}name" % _namespace)
                name.text = y
                member.append(name)
                member.append(_py2xml(x[y]))
                struct.append(member)
            val.append(struct)
        return val

def xml2py(params):
    namespace = 'jabber:iq:rpc'
    vals = []
    for param in params.findall('{%s}param' % namespace):
        vals.append(_xml2py(param.find('{%s}value' % namespace)))
    return vals

def _xml2py(value):
    namespace = 'jabber:iq:rpc'
    find_value = value.find
    if find_value('{%s}nil' % namespace) is not None:
        return None
    if find_value('{%s}i4' % namespace) is not None:
        return int(find_value('{%s}i4' % namespace).text)
    if find_value('{%s}int' % namespace) is not None:
        return int(find_value('{%s}int' % namespace).text)
    if find_value('{%s}boolean' % namespace) is not None:
        return bool(int(find_value('{%s}boolean' % namespace).text))
    if find_value('{%s}string' % namespace) is not None:
        return find_value('{%s}string' % namespace).text
    if find_value('{%s}double' % namespace) is not None:
        return float(find_value('{%s}double' % namespace).text)
    if find_value('{%s}base64' % namespace) is not None:
        return rpcbase64(find_value('{%s}base64' % namespace).text.encode())
    if find_value('{%s}Base64' % namespace) is not None:
        # Older versions of XEP-0009 used Base64
        return rpcbase64(find_value('{%s}Base64' % namespace).text.encode())
    if find_value('{%s}dateTime.iso8601' % namespace) is not None:
        return rpctime(find_value('{%s}dateTime.iso8601' % namespace).text)
    if find_value('{%s}struct' % namespace) is not None:
        struct = {}
        for member in find_value('{%s}struct' % namespace).findall('{%s}member' % namespace):
            struct[member.find('{%s}name' % namespace).text] = _xml2py(member.find('{%s}value' % namespace))
        return struct
    if find_value('{%s}array' % namespace) is not None:
        array = []
        for val in find_value('{%s}array' % namespace).find('{%s}data' % namespace).findall('{%s}value' % namespace):
            array.append(_xml2py(val))
        return array
    raise ValueError()


class rpcbase64:

    def __init__(self, data):
        #base 64 encoded string
        self.data = data

    def decode(self):
        return base64.b64decode(self.data)

    def __str__(self):
        return self.decode().decode()

    def encoded(self):
        return self.data.decode()



class rpctime:

    def __init__(self,data=None):
        #assume string data is in iso format YYYYMMDDTHH:MM:SS
        if type(data) is str:
            self.timestamp = time.strptime(data,"%Y%m%dT%H:%M:%S")
        elif type(data) is time.struct_time:
            self.timestamp = data
        elif data is None:
            self.timestamp = time.gmtime()
        else:
            raise ValueError()

    def iso8601(self):
        #return a iso8601 string
        return time.strftime("%Y%m%dT%H:%M:%S",self.timestamp)

    def __str__(self):
        return self.iso8601()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Error
from slixmpp.xmlstream import ElementBase, StanzaBase


class Enable(StanzaBase):
    name = 'enable'
    namespace = 'urn:xmpp:sm:3'
    interfaces = {'max', 'resume'}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_resume(self):
        return self._get_attr('resume', 'false').lower() in ('true', '1')

    def set_resume(self, val):
        self._del_attr('resume')
        self._set_attr('resume', 'true' if val else 'false')


class Enabled(StanzaBase):
    name = 'enabled'
    namespace = 'urn:xmpp:sm:3'
    interfaces = {'id', 'location', 'max', 'resume'}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_resume(self):
        return self._get_attr('resume', 'false').lower() in ('true', '1')

    def set_resume(self, val):
        self._del_attr('resume')
        self._set_attr('resume', 'true' if val else 'false')


class Resume(StanzaBase):
    name = 'resume'
    namespace = 'urn:xmpp:sm:3'
    interfaces = {'h', 'previd'}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_h(self):
        h = self._get_attr('h', None)
        if h:
            return int(h)
        return None

    def set_h(self, val):
        self._set_attr('h', str(val))


class Resumed(StanzaBase):
    name = 'resumed'
    namespace = 'urn:xmpp:sm:3'
    interfaces = {'h', 'previd'}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_h(self):
        h = self._get_attr('h', None)
        if h:
            return int(h)
        return None

    def set_h(self, val):
        self._set_attr('h', str(val))


class Failed(StanzaBase, Error):
    name = 'failed'
    namespace = 'urn:xmpp:sm:3'
    interfaces = set()

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()


class StreamManagement(ElementBase):
    name = 'sm'
    namespace = 'urn:xmpp:sm:3'
    plugin_attrib = name
    interfaces = {'required', 'optional'}

    def get_required(self):
        return self.xml.find('{%s}required' % self.namespace) is not None

    def set_required(self, val):
        self.del_required()
        if val:
            self._set_sub_text('required', '', keep=True)

    def del_required(self):
        self._del_sub('required')

    def get_optional(self):
        return self.xml.find('{%s}optional' % self.namespace) is not None

    def set_optional(self, val):
        self.del_optional()
        if val:
            self._set_sub_text('optional', '', keep=True)

    def del_optional(self):
        self._del_sub('optional')


class RequestAck(StanzaBase):
    name = 'r'
    namespace = 'urn:xmpp:sm:3'
    interfaces = set()

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()


class Ack(StanzaBase):
    name = 'a'
    namespace = 'urn:xmpp:sm:3'
    interfaces = {'h'}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

    def get_h(self):
        h = self._get_attr('h', None)
        if h:
            return int(h)
        return None

    def set_h(self, val):
        self._set_attr('h', str(val))




# slixmpp: The Slick XMPP Library
# Implementation of HTTP over XMPP transport
# http://xmpp.org/extensions/xep-0332.html
# Copyright (C) 2015 Riptide IO, sangeeth@riptideio.com
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class HTTPRequest(ElementBase):

    """
    All HTTP communication is done using the `Request`/`Response` paradigm.
    Each HTTP Request is made sending an `iq` stanza containing a `req`
    element to the server. Each `iq` stanza sent is of type `set`.

    Examples:
    ::

        <iq type='set' from='a@b.com/browser' to='x@y.com' id='1'>
            <req xmlns='urn:xmpp:http'
                 method='GET'
                 resource='/api/users'
                 version='1.1'>
                <headers xmlns='http://jabber.org/protocol/shim'>
                    <header name='Host'>b.com</header>
                </headers>
            </req>
        </iq>

        <iq type='set' from='a@b.com/browser' to='x@y.com' id='2'>
            <req xmlns='urn:xmpp:http'
                 method='PUT'
                 resource='/api/users'
                 version='1.1'>
                <headers xmlns='http://jabber.org/protocol/shim'>
                    <header name='Host'>b.com</header>
                    <header name='Content-Type'>text/html</header>
                    <header name='Content-Length'>...</header>
                </headers>
                <data>
                    <text>...</text>
                </data>
            </req>
        </iq>
    """

    name = 'request'
    namespace = 'urn:xmpp:http'
    interfaces = {'method', 'resource', 'version'}
    plugin_attrib = 'http-req'

    def get_method(self):
        return self._get_attr('method', None)

    def set_method(self, method):
        self._set_attr('method', method)

    def get_resource(self):
        return self._get_attr('resource', None)

    def set_resource(self, resource):
        self._set_attr('resource', resource)

    def get_version(self):
        return self._get_attr('version', None)

    def set_version(self, version='1.1'):
        self._set_attr('version', version)



import unittest
from slixmpp.test import SlixTest
from slixmpp import JID


class TestMIXPAM(SlixTest):

    def setUp(self):
        self.stream_start(plugins=['xep_0405'])

    def testGetRosterEmpty(self):
        """Test requesting an empty annotated roster"""

        fut = self.xmpp.wrap(self.xmpp['xep_0405'].get_mix_roster())

        self.wait_()
        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster">
                <annotate xmlns='urn:xmpp:mix:roster:0' />
            </query>
          </iq>
        """)

        self.recv("""
          <iq type="result" id="1"
              to="tester@localhost" />
        """)

        result = self.run_coro(fut)

    def testGetRoster(self):
        """Test requesting an annotated roster"""

        fut = self.xmpp.wrap(self.xmpp['xep_0405'].get_mix_roster())

        self.wait_()
        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster">
                <annotate xmlns='urn:xmpp:mix:roster:0' />
            </query>
          </iq>
        """)

        self.recv("""
          <iq type="result" id="1" to="tester@localhost">
            <query xmlns="jabber:iq:roster">
              <item jid='romeo@example.net'/>
              <item jid='juliet@example.net'/>
              <item jid='balcony@example.net'>
                <channel xmlns='urn:xmpp:mix:roster:0'
                         participant-id='123456'/>
              </item>
            </query>
          </iq>
        """)

        self.wait_()
        contacts, channels = fut.result()
        self.assertEqual(len(contacts), 2)
        self.assertEqual(contacts[0]['jid'], 'romeo@example.net')
        self.assertEqual(contacts[1]['jid'], 'juliet@example.net')
        self.assertEqual(len(channels), 1)
        self.assertEqual(channels[0]['jid'], 'balcony@example.net')
        self.assertEqual(
            channels[0]['channel']['participant-id'],
            '123456'
        )

    def testClientJoin(self):
        """Test a client join"""

        fut = self.xmpp.wrap(self.xmpp['xep_0405'].join_channel(
            JID('coven@mix.shakespeare.example'),
            'toto',
        ))
        self.send("""
            <iq type='set' to='tester@localhost' id='1'>
              <client-join xmlns='urn:xmpp:mix:pam:2'
                           channel='coven@mix.shakespeare.example'>
                <join xmlns='urn:xmpp:mix:core:1'>
                  <nick>toto</nick>
                  <subscribe node='urn:xmpp:mix:nodes:messages'/>
                  <subscribe node='urn:xmpp:mix:nodes:participants'/>
                  <subscribe node='urn:xmpp:mix:nodes:info'/>
                </join>
              </client-join>
            </iq>
        """)
        self.recv("""
            <iq type='result'
                from='tester@localhost'
                to='tester@localhost/resource'
                id='1'>
                  <client-join xmlns='urn:xmpp:mix:pam:2'>
                    <join xmlns='urn:xmpp:mix:core:1'
                          jid='123456#coven@mix.shakespeare.example'>
                      <subscribe node='urn:xmpp:mix:nodes:messages'/>
                      <subscribe node='urn:xmpp:mix:nodes:participants'/>
                      <subscribe node='urn:xmpp:mix:nodes:info'/>
                   </join>
                 </client-join>
            </iq>
        """)
        self.wait_()
        self.assertEqual(fut.result(), set())

    def testClientJoinNotAllNodes(self):
        """Test a client join where one of the nodes is rejected"""

        fut = self.xmpp.wrap(self.xmpp['xep_0405'].join_channel(
            JID('coven@mix.shakespeare.example'),
            'toto',
        ))
        self.send("""
            <iq type='set' to='tester@localhost' id='1'>
              <client-join xmlns='urn:xmpp:mix:pam:2'
                           channel='coven@mix.shakespeare.example'>
                <join xmlns='urn:xmpp:mix:core:1'>
                  <nick>toto</nick>
                  <subscribe node='urn:xmpp:mix:nodes:messages'/>
                  <subscribe node='urn:xmpp:mix:nodes:participants'/>
                  <subscribe node='urn:xmpp:mix:nodes:info'/>
                </join>
              </client-join>
            </iq>
        """)
        self.recv("""
            <iq type='result'
                from='tester@localhost'
                to='tester@localhost/resource'
                id='1'>
                  <client-join xmlns='urn:xmpp:mix:pam:2'>
                    <join xmlns='urn:xmpp:mix:core:1'
                          jid='123456#coven@mix.shakespeare.example'>
                      <subscribe node='urn:xmpp:mix:nodes:messages'/>
                      <subscribe node='urn:xmpp:mix:nodes:participants'/>
                   </join>
                 </client-join>
            </iq>
        """)
        self.wait_()
        self.assertEqual(fut.result(), {'urn:xmpp:mix:nodes:info'})

    def testClientLeave(self):
        """Test a client leave"""

        fut = self.xmpp.wrap(self.xmpp['xep_0405'].leave_channel(
            JID('coven@mix.shakespeare.example'),
        ))
        self.send("""
            <iq type='set'
                to='tester@localhost'
                id='1'>
              <client-leave xmlns='urn:xmpp:mix:pam:2'
                            channel='coven@mix.shakespeare.example'>
                <leave xmlns='urn:xmpp:mix:core:1'/>
              </client-leave>
            </iq>
        """)
        self.recv("""
            <iq type='result'
                from='tester@localhost'
                to='tester@localhost/resource'
                id='1'>
              <client-leave xmlns='urn:xmpp:mix:pam:2'
                            channel='coven@mix.shakespeare.example'>
                <leave xmlns='urn:xmpp:mix:core:1'/>
              </client-leave>
            </iq>
        """)

        self.assertEqual(fut.done(), True)
        self.assertEqual(fut.exception(), None)


suite = unittest.TestLoader().loadTestsFromTestCase(TestMIXPAM)




# slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
import logging

from typing import Iterable, Tuple, Optional

from slixmpp import JID, Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0353 import stanza, Propose, Retract, Accept, Proceed, Reject

log = logging.getLogger(__name__)

class XEP_0353(BasePlugin):

    name = 'xep_0353'
    description = 'XEP-0353: Jingle Message Initiation'
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, Propose)
        register_stanza_plugin(Message, Retract)
        register_stanza_plugin(Message, Accept)
        register_stanza_plugin(Message, Proceed)
        register_stanza_plugin(Message, Reject)

        self.xmpp.register_handler(
            Callback('Indicating Intent to Start a Session',
                StanzaPath('message/jingle_propose'),
                self._handle_propose))
        self.xmpp.register_handler(
            Callback('Disavowing Intent to Start a Session',
                StanzaPath('message/jingle_retract'),
                self._handle_retract))
        self.xmpp.register_handler(
            Callback('Accepting Intent to Start a Session',
                StanzaPath('message/jingle_accept'),
                self._handle_accept))
        self.xmpp.register_handler(
            Callback('Proceed',
                StanzaPath('message/jingle_proceed'),
                self._handle_proceed))
        self.xmpp.register_handler(
            Callback('Rejecting Intent to Start a Session',
                StanzaPath('message/jingle_reject'),
                self._handle_reject))

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(stanza.JingleMessage.namespace)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.JingleMessage.namespace)

    def _handle_propose(self, message):
        self.xmpp.event('jingle_message_propose', message)

    def _handle_retract(self, message):
        self.xmpp.event('jingle_message_retract', message)

    def _handle_accept(self, message):
        self.xmpp.event('jingle_message_accept', message)

    def _handle_proceed(self, message):
        self.xmpp.event('jingle_message_proceed', message)

    def _handle_reject(self, message):
        self.xmpp.event('jingle_message_reject', message)

    def propose(self, mto: JID, sid: str, descriptions: Iterable[Tuple[str, str]], *, mfrom: Optional[JID] = None):
        msg = self.xmpp.make_message(mto, mfrom=mfrom)
        msg['jingle_propose']['id'] = sid
        msg['jingle_propose']['descriptions'] = descriptions
        msg.send()

    def retract(self, mto: JID, sid: str, *, mfrom: Optional[JID] = None):
        msg = self.xmpp.make_message(mto, mfrom=mfrom)
        msg['jingle_retract']['id'] = sid
        msg.send()

    def accept(self, mto: JID, sid: str, *, mfrom: Optional[JID] = None):
        msg = self.xmpp.make_message(mto, mfrom=mfrom)
        msg['jingle_accept']['id'] = sid
        msg.send()

    def proceed(self, mto: JID, sid: str, *, mfrom: Optional[JID] = None):
        msg = self.xmpp.make_message(mto, mfrom=mfrom)
        msg['jingle_proceed']['id'] = sid
        msg.send()

    def reject(self, mto: JID, sid: str, *, mfrom: Optional[JID] = None):
        msg = self.xmpp.make_message(mto, mfrom=mfrom)
        msg['jingle_reject']['id'] = sid
        msg.send()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.stanza import StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.matcher import MatchXPath
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.features.feature_starttls import stanza

from typing import ClassVar, Set


log = logging.getLogger(__name__)


class FeatureSTARTTLS(BasePlugin):

    name = 'feature_starttls'
    description = 'RFC 6120: Stream Feature: STARTTLS'
    dependencies: ClassVar[Set[str]] = set()
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
                CoroutineCallback('STARTTLS Proceed',
                        MatchXPath(stanza.Proceed.tag_name()),
                        self._handle_starttls_proceed,
                        instream=True))
        self.xmpp.register_feature('starttls',
                self._handle_starttls,
                restart=True,
                order=self.config.get('order', 0))

        self.xmpp.register_stanza(stanza.Proceed)
        self.xmpp.register_stanza(stanza.Failure)
        register_stanza_plugin(StreamFeatures, stanza.STARTTLS)

    def _handle_starttls(self, features):
        """
        Handle notification that the server supports TLS.

        Arguments:
            features -- The stream:features element.
        """
        if 'starttls' in self.xmpp.features:
            # We have already negotiated TLS, but the server is
            # offering it again, against spec.
            return False
        elif self.xmpp.disable_starttls:
            return False
        else:
            self.xmpp.send(stanza.STARTTLS())
            return True

    async def _handle_starttls_proceed(self, proceed):
        """Restart the XML stream when TLS is accepted."""
        log.debug("Starting TLS")
        if await self.xmpp.start_tls():
            self.xmpp.features.add('starttls')




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)


NS = 'urn:xmpp:message-retract:1'


class Retract(ElementBase):
    namespace = NS
    name = 'retract'
    plugin_attrib = 'retract'
    interfaces = {'reason', 'id'}
    sub_interfaces = {'reason'}


class Retracted(ElementBase):
    namespace = NS
    name = 'retracted'
    plugin_attrib = 'retracted'
    interfaces = {'stamp', 'id', 'reason'}
    sub_interfaces = {'reason'}


def register_plugins():
    register_stanza_plugin(Message, Retract)
    register_stanza_plugin(Message, Retracted)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0258 import stanza
from slixmpp.plugins.xep_0258.stanza import SecurityLabel, Label
from slixmpp.plugins.xep_0258.stanza import DisplayMarking, EquivalentLabel
from slixmpp.plugins.xep_0258.stanza import ESSLabel, Catalog, CatalogItem
from slixmpp.plugins.xep_0258.security_labels import XEP_0258


register_plugin(XEP_0258)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import Set, Iterable
from slixmpp.xmlstream import ElementBase
try:
    from emoji import is_emoji
except ImportError:
    def is_emoji(*args, **kwargs) -> bool:
        return True


NS = 'urn:xmpp:reactions:0'

class Reactions(ElementBase):
    name = 'reactions'
    plugin_attrib = 'reactions'
    namespace = NS
    interfaces = {'id', 'values'}

    def get_values(self, *, all_chars=False) -> Set[str]:
        """"Get all reactions as str"""
        reactions = set()
        for reaction in self:
            value = reaction['value']
            if all_chars:
                reactions.add(reaction['value'])
            elif is_emoji(value):
                reactions.add(reaction['value'])
        return reactions

    def set_values(self, values: Iterable[str], *, all_chars=False):
        """"Set all reactions as str"""
        for element in self.xml.findall('reaction'):
            self.xml.remove(element)
        for reaction_txt in values:
            reaction = Reaction()
            reaction.set_value(reaction_txt, all_chars=all_chars)
            self.append(reaction)


class Reaction(ElementBase):
    name = 'reaction'
    namespace = NS
    interfaces = {'value'}

    def get_value(self) -> str:
        return self.xml.text

    def set_value(self, value: str, *, all_chars=False):
        if not all_chars and not is_emoji(value):
            raise ValueError("%s is not a valid emoji" % value)
        self.xml.text = value




import unittest
from datetime import datetime
from slixmpp.test import SlixTest
from slixmpp import JID


class TestMAM(SlixTest):

    def setUp(self):
        self.stream_start(plugins=['xep_0313'])

    def testRetrieveSimple(self):
        """Test requesting MAM messages without RSM"""

        msgs = []

        async def test():
            iq = await self.xmpp['xep_0313'].retrieve()
            for message in iq['mam']['results']:
                msgs.append(message)

        fut = self.xmpp.wrap(test())
        self.wait_()
        self.send("""
            <iq type='set' id='1'>
              <query xmlns='urn:xmpp:mam:2' queryid='1' />
            </iq>
        """)

        self.recv("""
            <message id='abc' to='tester@localhost/resource'>
              <result xmlns='urn:xmpp:mam:2' queryid='1'
                      id='28482-98726-73623'>
                <forwarded xmlns='urn:xmpp:forward:0'>
                  <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:25Z'/>
                  <message xmlns='jabber:client' from="witch@shakespeare.lit"
                           to="tester@localhost">
                    <body>Hail to thee</body>
                  </message>
                </forwarded>
              </result>
            </message>
        """)

        self.recv("""
          <iq type="result" id="1" to="tester@localhost">
            <fin xmlns="urn:xmpp:mam:2">
              <first index='0'>28482-98726-73623</first>
              <last>28482-98726-73623</last>
            </fin>
          </iq>
        """)

        self.run_coro(fut)
        self.assertEqual(
            msgs[0]['mam_result']['forwarded']['message']['body'],
            "Hail to thee",
        )
        self.assertEqual(len(msgs),1)

    def testRetrieveRSM(self):
        """Test requesting MAM messages with RSM"""

        msgs = []

        async def test():
            iterator = self.xmpp['xep_0313'].retrieve(
                with_jid=JID('toto@titi'),
                start='2010-06-07T00:00:00Z',
                iterator=True,
            )
            async for page in iterator:
                for message in page['mam']['results']:
                    msgs.append(message)

        fut = self.xmpp.wrap(test())
        self.wait_()
        self.send("""
            <iq type='set' id='2'>
              <query xmlns='urn:xmpp:mam:2' queryid='2'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field var='FORM_TYPE' type='hidden'>
                    <value>urn:xmpp:mam:2</value>
                  </field>
                  <field var='with'>
                    <value>toto@titi</value>
                  </field>
                  <field var='start'>
                    <value>2010-06-07T00:00:00Z</value>
                  </field>
                </x>
                <set xmlns="http://jabber.org/protocol/rsm">
                  <max>10</max>
                </set>
              </query>
            </iq>
        """)

        self.recv("""
            <message id='abc' to='tester@localhost/resource'>
              <result xmlns='urn:xmpp:mam:2' queryid='2'
                      id='28482-98726-73623'>
                <forwarded xmlns='urn:xmpp:forward:0'>
                  <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:25Z'/>
                  <message xmlns='jabber:client' from="witch@shakespeare.lit"
                           to="tester@localhost">
                    <body>Hail to thee</body>
                  </message>
                </forwarded>
              </result>
            </message>
        """)

        self.recv("""
          <iq type="result" id="2" to="tester@localhost">
            <fin xmlns="urn:xmpp:mam:2">
              <set xmlns='http://jabber.org/protocol/rsm'>
                <first index='0'>28482-98726-73623</first>
                 <last>28482-98726-73623</last>
                 <count>2</count>
              </set>
            </fin>
          </iq>
        """)

        self.send("""
            <iq type='set' id='3'>
              <query xmlns='urn:xmpp:mam:2' queryid='3'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field var='FORM_TYPE' type='hidden'>
                    <value>urn:xmpp:mam:2</value>
                  </field>
                  <field var='with'>
                    <value>toto@titi</value>
                  </field>
                  <field var='start'>
                    <value>2010-06-07T00:00:00Z</value>
                  </field>
                </x>
                <set xmlns="http://jabber.org/protocol/rsm">
                  <max>10</max>
                 <after>28482-98726-73623</after>
                </set>
              </query>
            </iq>
        """)

        self.recv("""
            <message id='abc' to='tester@localhost/resource'>
              <result xmlns='urn:xmpp:mam:2' queryid='3'
                      id='28482-98726-73624'>
                <forwarded xmlns='urn:xmpp:forward:0'>
                  <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:26Z'/>
                  <message xmlns='jabber:client' from="witch@shakespeare.lit"
                           to="tester@localhost">
                    <body>Hi Y'all</body>
                  </message>
                </forwarded>
              </result>
            </message>
        """)

        self.recv("""
          <iq type="result" id="3" to="tester@localhost">
            <fin xmlns="urn:xmpp:mam:2">
              <set xmlns='http://jabber.org/protocol/rsm'>
                <first index='1'>28482-98726-73624</first>
                 <last>28482-98726-73624</last>
                 <count>2</count>
              </set>
            </fin>
          </iq>
        """)

        self.run_coro(fut)
        self.assertEqual(
            msgs[0]['mam_result']['forwarded']['message']['body'],
            "Hail to thee",
        )
        self.assertEqual(
            msgs[1]['mam_result']['forwarded']['message']['body'],
            "Hi Y'all",
        )
        self.assertEqual(len(msgs), 2)

    def testIterate(self):
        """Test iterating over MAM messages with RSM"""

        msgs = []

        async def test():
            iterator = self.xmpp['xep_0313'].iterate(
                with_jid=JID('toto@titi'),
                start='2010-06-07T00:00:00Z',
            )
            async for message in iterator:
                msgs.append(message)

        fut = self.xmpp.wrap(test())
        self.wait_()
        self.send("""
            <iq type='set' id='2'>
              <query xmlns='urn:xmpp:mam:2' queryid='2'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field var='FORM_TYPE' type='hidden'>
                    <value>urn:xmpp:mam:2</value>
                  </field>
                  <field var='with'>
                    <value>toto@titi</value>
                  </field>
                  <field var='start'>
                    <value>2010-06-07T00:00:00Z</value>
                  </field>
                </x>
                <set xmlns="http://jabber.org/protocol/rsm">
                  <max>10</max>
                </set>
              </query>
            </iq>
        """)

        self.recv("""
            <message id='abc' to='tester@localhost/resource'>
              <result xmlns='urn:xmpp:mam:2' queryid='2'
                             id='28482-98726-73623'>
                <forwarded xmlns='urn:xmpp:forward:0'>
                  <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:25Z'/>
                  <message xmlns='jabber:client' from="witch@shakespeare.lit"
                           to="tester@localhost">
                    <body>Hail to thee</body>
                  </message>
                </forwarded>
              </result>
            </message>
        """)

        self.recv("""
          <iq type="result" id="2" to="tester@localhost">
            <fin xmlns="urn:xmpp:mam:2">
              <set xmlns='http://jabber.org/protocol/rsm'>
                <first index='0'>28482-98726-73623</first>
                 <last>28482-98726-73623</last>
                 <count>2</count>
              </set>
            </fin>
          </iq>
        """)

        self.send("""
            <iq type='set' id='3'>
              <query xmlns='urn:xmpp:mam:2' queryid='3'>
                <x xmlns='jabber:x:data' type='submit'>
                  <field var='FORM_TYPE' type='hidden'>
                    <value>urn:xmpp:mam:2</value>
                  </field>
                  <field var='with'>
                    <value>toto@titi</value>
                  </field>
                  <field var='start'>
                    <value>2010-06-07T00:00:00Z</value>
                  </field>
                </x>
                <set xmlns="http://jabber.org/protocol/rsm">
                  <max>10</max>
                 <after>28482-98726-73623</after>
                </set>
              </query>
            </iq>
        """)

        self.recv("""
            <message id='abc' to='tester@localhost/resource'>
              <result xmlns='urn:xmpp:mam:2' queryid='3'
                      id='28482-98726-73624'>
                <forwarded xmlns='urn:xmpp:forward:0'>
                  <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:26Z'/>
                  <message xmlns='jabber:client' from="witch@shakespeare.lit"
                           to="tester@localhost">
                    <body>Hi Y'all</body>
                  </message>
                </forwarded>
              </result>
            </message>
        """)

        self.recv("""
          <iq type="result" id="3" to="tester@localhost">
            <fin xmlns="urn:xmpp:mam:2">
              <set xmlns='http://jabber.org/protocol/rsm'>
                <first index='1'>28482-98726-73624</first>
                 <last>28482-98726-73624</last>
                 <count>2</count>
              </set>
            </fin>
          </iq>
        """)

        self.run_coro(fut)
        self.assertEqual(
            msgs[0]['mam_result']['forwarded']['message']['body'],
            "Hail to thee",
        )
        self.assertEqual(
            msgs[1]['mam_result']['forwarded']['message']['body'],
            "Hi Y'all",
        )
        self.assertEqual(len(msgs), 2)

    def test_get_metadata(self):
        """Test a MAM metadata retrieval"""
        fut = self.xmpp.wrap(
            self.xmpp.plugin['xep_0313'].get_archive_metadata()
        )
        self.wait_()
        self.send("""
            <iq type='get' id='1'>
              <metadata xmlns='urn:xmpp:mam:2'/>
            </iq>
        """)
        self.recv("""
            <iq type='result' id='1'>
              <metadata xmlns='urn:xmpp:mam:2'>
                <start id='YWxwaGEg' timestamp='2008-08-22T21:09:04Z' />
                <end id='b21lZ2Eg' timestamp='2020-04-20T14:34:21Z' />
              </metadata>
            </iq>
        """)
        self.run_coro(fut)
        result = fut.result()
        self.assertEqual(result['mam_metadata']['start']['id'], "YWxwaGEg")
        self.assertEqual(
            result['mam_metadata']['start']['timestamp'],
            datetime.fromisoformat('2008-08-22T21:09:04+00:00')
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestMAM)



import unittest
from slixmpp.test import SlixTest


class TestPrivatePEP(SlixTest):

    def testConfigureNode(self):
        self.stream_start(mode='client', plugins=['xep_0223'])

        self.xmpp.plugin['xep_0223'].configure(node="toto")
        self.send("""
<iq type="set" id="1">
  <pubsub xmlns='http://jabber.org/protocol/pubsub#owner'>
    <configure node='toto'>
      <x xmlns='jabber:x:data' type='submit'>
        <field var='FORM_TYPE' type='hidden'>
          <value>http://jabber.org/protocol/pubsub#node_config</value>
        </field>
        <field var='pubsub#persist_items'><value>1</value></field>
        <field var='pubsub#access_model'><value>whitelist</value></field>
      </x>
    </configure>
  </pubsub>
</iq>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestPrivatePEP)



import unittest
from slixmpp import Message, Iq, JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0424 import stanza as stanza424
from slixmpp.plugins.xep_0425 import stanza


class TestModeration(SlixTest):

    def setUp(self):
        stanza424.register_plugins()
        stanza.register_plugins()

    def testModerate(self):
        iq = Iq()
        iq['type'] = 'set'
        iq['id'] = 'a'
        iq['moderate']['id'] = 'some-id'
        iq['moderate'].enable('retract')
        iq['moderate']['reason'] = 'R'

        self.check(iq, """
<iq type='set' id='a'>
  <moderate xmlns='urn:xmpp:message-moderate:1' id='some-id'>
    <retract xmlns='urn:xmpp:message-retract:1'/>
    <reason>R</reason>
  </moderate>
</iq>
        """)

    def testModerated(self):
        message = Message()
        message['retract']['id'] = 'some-id'
        message['retract']['moderated']['by'] = JID('toto@titi')
        message['retract']['moderated']['occupant-id']['id'] = 'oc-id'
        message['retract']['reason'] = 'R'

        self.check(message, """
<message>
  <retract id='some-id' xmlns='urn:xmpp:message-retract:1'>
    <moderated by='toto@titi' xmlns='urn:xmpp:message-moderate:1'>
      <occupant-id xmlns="urn:xmpp:occupant-id:0" id="oc-id" />
    </moderated>
    <reason>R</reason>
  </retract>
</message>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestModeration)



import unittest
from slixmpp import ET
from slixmpp.test.integration import SlixIntegration
from slixmpp.plugins.xep_0048.stanza import Bookmarks


class TestPrivateStorage(SlixIntegration):
    async def asyncSetUp(self):
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.register_plugins(['xep_0048', 'xep_0049'])
        await self.connect_clients()

    async def test_privatestorage(self):
        """Check we can set, get, and delete private in xml storage"""
        # Set a bookmark using private storage
        el = Bookmarks()
        el.add_conference('test@example.com', 'toto')
        await self.clients[0]['xep_0049'].store(
            el,
        )
        result = await self.clients[0]['xep_0049'].retrieve('bookmarks')

        self.assertEqual(result['private']['bookmarks'], el)

        # Purge bookmarks
        await self.clients[0]['xep_0049'].store(
            Bookmarks(),
        )
        result = await self.clients[0]['xep_0049'].retrieve('bookmarks')
        self.assertEqual(result['private']['bookmarks'], Bookmarks())


suite = unittest.TestLoader().loadTestsFromTestCase(TestPrivateStorage)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp


class EchoBot(slixmpp.ClientXMPP):

    """
    A simple Slixmpp bot that will echo messages it
    receives, along with a short thank you message.
    """

    def __init__(self, jid, password):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.start)

        # The message event is triggered whenever a message
        # stanza is received. Be aware that that includes
        # MUC messages and error messages.
        self.add_event_handler("message", self.message)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        await self.get_roster()

    def message(self, msg):
        """
        Process incoming message stanzas. Be aware that this also
        includes MUC messages and error messages. It is usually
        a good idea to check the messages's type before processing
        or sending replies.

        Arguments:
            msg -- The received message stanza. See the documentation
                   for stanza objects and the Message stanza to see
                   how it may be used.
        """
        if msg['type'] in ('chat', 'normal'):
            msg.reply("Thanks for sending\n%(body)s" % msg).send()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser(description=EchoBot.__doc__)

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    # Setup the EchoBot and register plugins. Note that while plugins may
    # have interdependencies, the order in which you register them does
    # not matter.
    xmpp = EchoBot(args.jid, args.password)
    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0004') # Data Forms
    xmpp.register_plugin('xep_0060') # PubSub
    xmpp.register_plugin('xep_0199') # XMPP Ping

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()



from slixmpp.xmlstream import ElementBase

NAMESPACE = "urn:xmpp:reference:0"


class Reference(ElementBase):
    name = plugin_attrib = "reference"
    namespace = NAMESPACE
    interfaces = {"type", "uri", "id", "begin", "end"}



""" A wrapper for the 'gpg' command::

Portions of this module are derived from A.M. Kuchling's well-designed
GPG.py, using Richard Jones' updated version 1.3, which can be found
in the pycrypto CVS repository on Sourceforge:

http://pycrypto.cvs.sourceforge.net/viewvc/pycrypto/gpg/GPG.py

This module is *not* forward-compatible with amk's; some of the
old interface has changed.  For instance, since I've added decrypt
functionality, I elected to initialize with a 'gnupghome' argument
instead of 'keyring', so that gpg can find both the public and secret
keyrings.  I've also altered some of the returned objects in order for
the caller to not have to know as much about the internals of the
result classes.

While the rest of ISconf is released under the GPL, I am releasing
this single file under the same terms that A.M. Kuchling used for
pycrypto.

Steve Traugott, stevegt@terraluna.org
Thu Jun 23 21:27:20 PDT 2005

This version of the module has been modified from Steve Traugott's version
(see http://trac.t7a.org/isconf/browser/trunk/lib/python/isconf/GPG.py) by
Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
the previous versions.

Modifications Copyright (C) 2008-2012 Vinay Sajip. All rights reserved.

A unittest harness (test_gnupg.py) has also been added.
"""
import locale

__version__ = "0.2.9"
__author__ = "Vinay Sajip"
__date__  = "$29-Mar-2012 21:12:58$"

try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO

import codecs
import locale
import logging
import os
import socket
from subprocess import Popen
from subprocess import PIPE
import sys
import threading

try:
    import logging.NullHandler as NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def handle(self, record):
            pass
try:
    unicode
    _py3k = False
except NameError:
    _py3k = True

logger = logging.getLogger(__name__)
if not logger.handlers:
    logger.addHandler(NullHandler())

def _copy_data(instream, outstream):
    # Copy one stream to another
    sent = 0
    if hasattr(sys.stdin, 'encoding'):
        enc = sys.stdin.encoding
    else:
        enc = 'ascii'
    while True:
        data = instream.read(1024)
        if len(data) == 0:
            break
        sent += len(data)
        logger.debug("sending chunk (%d): %r", sent, data[:256])
        try:
            outstream.write(data)
        except UnicodeError:
            outstream.write(data.encode(enc))
        except:
            # Can sometimes get 'broken pipe' errors even when the data has all
            # been sent
            logger.exception('Error sending data')
            break
    try:
        outstream.close()
    except IOError:
        logger.warning('Exception occurred while closing: ignored', exc_info=1)
    logger.debug("closed output, %d bytes sent", sent)

def _threaded_copy_data(instream, outstream):
    wr = threading.Thread(target=_copy_data, args=(instream, outstream))
    wr.setDaemon(True)
    logger.debug('data copier: %r, %r, %r', wr, instream, outstream)
    wr.start()
    return wr

def _write_passphrase(stream, passphrase, encoding):
    passphrase = '%s\n' % passphrase
    passphrase = passphrase.encode(encoding)
    stream.write(passphrase)
    logger.debug("Wrote passphrase: %r", passphrase)

def _is_sequence(instance):
    return isinstance(instance,list) or isinstance(instance,tuple)

def _make_binary_stream(s, encoding):
    try:
        if _py3k:
            if isinstance(s, str):
                s = s.encode(encoding)
        else:
            if type(s) is not str:
                s = s.encode(encoding)
        from io import BytesIO
        rv = BytesIO(s)
    except ImportError:
        rv = StringIO(s)
    return rv

class Verify(object):
    "Handle status messages for --verify"

    def __init__(self, gpg):
        self.gpg = gpg
        self.valid = False
        self.fingerprint = self.creation_date = self.timestamp = None
        self.signature_id = self.key_id = None
        self.username = None

    def __nonzero__(self):
        return self.valid

    __bool__ = __nonzero__

    def handle_status(self, key, value):
        if key in ("TRUST_UNDEFINED", "TRUST_NEVER", "TRUST_MARGINAL",
                   "TRUST_FULLY", "TRUST_ULTIMATE", "RSA_OR_IDEA", "NODATA",
                   "IMPORT_RES", "PLAINTEXT", "PLAINTEXT_LENGTH",
                   "POLICY_URL", "DECRYPTION_INFO", "DECRYPTION_OKAY", "IMPORTED"):
            pass
        elif key == "BADSIG":
            self.valid = False
            self.status = 'signature bad'
            self.key_id, self.username = value.split(None, 1)
        elif key == "GOODSIG":
            self.valid = True
            self.status = 'signature good'
            self.key_id, self.username = value.split(None, 1)
        elif key == "VALIDSIG":
            (self.fingerprint,
             self.creation_date,
             self.sig_timestamp,
             self.expire_timestamp) = value.split()[:4]
            # may be different if signature is made with a subkey
            self.pubkey_fingerprint = value.split()[-1]
            self.status = 'signature valid'
        elif key == "SIG_ID":
            (self.signature_id,
             self.creation_date, self.timestamp) = value.split()
        elif key == "ERRSIG":
            self.valid = False
            (self.key_id,
             algo, hash_algo,
             cls,
             self.timestamp) = value.split()[:5]
            self.status = 'signature error'
        elif key == "DECRYPTION_FAILED":
            self.valid = False
            self.key_id = value
            self.status = 'decryption failed'
        elif key == "NO_PUBKEY":
            self.valid = False
            self.key_id = value
            self.status = 'no public key'
        elif key in ("KEYEXPIRED", "SIGEXPIRED"):
            # these are useless in verify, since they are spit out for any
            # pub/subkeys on the key, not just the one doing the signing.
            # if we want to check for signatures with expired key,
            # the relevant flag is EXPKEYSIG.
            pass
        elif key in ("EXPKEYSIG", "REVKEYSIG"):
            # signed with expired or revoked key
            self.valid = False
            self.key_id = value.split()[0]
            self.status = (('%s %s') % (key[:3], key[3:])).lower()
        else:
            raise ValueError("Unknown status message: %r" % key)

class ImportResult(object):
    "Handle status messages for --import"

    counts = '''count no_user_id imported imported_rsa unchanged
            n_uids n_subk n_sigs n_revoc sec_read sec_imported
            sec_dups not_imported'''.split()
    def __init__(self, gpg):
        self.gpg = gpg
        self.imported = []
        self.results = []
        self.fingerprints = []
        for result in self.counts:
            setattr(self, result, None)

    def __nonzero__(self):
        if self.not_imported: return False
        if not self.fingerprints: return False
        return True

    __bool__ = __nonzero__

    ok_reason = {
        '0': 'Not actually changed',
        '1': 'Entirely new key',
        '2': 'New user IDs',
        '4': 'New signatures',
        '8': 'New subkeys',
        '16': 'Contains private key',
    }

    problem_reason = {
        '0': 'No specific reason given',
        '1': 'Invalid Certificate',
        '2': 'Issuer Certificate missing',
        '3': 'Certificate Chain too long',
        '4': 'Error storing certificate',
    }

    def handle_status(self, key, value):
        if key == "IMPORTED":
            # this duplicates info we already see in import_ok & import_problem
            pass
        elif key == "NODATA":
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'No valid data found'})
        elif key == "IMPORT_OK":
            reason, fingerprint = value.split()
            reasons = []
            for code, text in list(self.ok_reason.items()):
                if int(reason) | int(code) == int(reason):
                    reasons.append(text)
            reasontext = '\n'.join(reasons) + "\n"
            self.results.append({'fingerprint': fingerprint,
                'ok': reason, 'text': reasontext})
            self.fingerprints.append(fingerprint)
        elif key == "IMPORT_PROBLEM":
            try:
                reason, fingerprint = value.split()
            except:
                reason = value
                fingerprint = '<unknown>'
            self.results.append({'fingerprint': fingerprint,
                'problem': reason, 'text': self.problem_reason[reason]})
        elif key == "IMPORT_RES":
            import_res = value.split()
            for i in range(len(self.counts)):
                setattr(self, self.counts[i], int(import_res[i]))
        elif key == "KEYEXPIRED":
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'Key expired'})
        elif key == "SIGEXPIRED":
            self.results.append({'fingerprint': None,
                'problem': '0', 'text': 'Signature expired'})
        else:
            raise ValueError("Unknown status message: %r" % key)

    def summary(self):
        l = []
        l.append('%d imported'%self.imported)
        if self.not_imported:
            l.append('%d not imported'%self.not_imported)
        return ', '.join(l)

class ListKeys(list):
    ''' Handle status messages for --list-keys.

        Handle pub and uid (relating the latter to the former).

        Don't care about (info from src/DETAILS):

        crt = X.509 certificate
        crs = X.509 certificate and private key available
        sub = subkey (secondary key)
        ssb = secret subkey (secondary key)
        uat = user attribute (same as user id except for field 10).
        sig = signature
        rev = revocation signature
        pkd = public key data (special field format, see below)
        grp = reserved for gpgsm
        rvk = revocation key
    '''
    def __init__(self, gpg):
        self.gpg = gpg
        self.curkey = None
        self.fingerprints = []
        self.uids = []

    def key(self, args):
        vars = ("""
            type trust length algo keyid date expires dummy ownertrust uid
        """).split()
        self.curkey = {}
        for i in range(len(vars)):
            self.curkey[vars[i]] = args[i]
        self.curkey['uids'] = []
        if self.curkey['uid']:
            self.curkey['uids'].append(self.curkey['uid'])
        del self.curkey['uid']
        self.append(self.curkey)

    pub = sec = key

    def fpr(self, args):
        self.curkey['fingerprint'] = args[9]
        self.fingerprints.append(args[9])

    def uid(self, args):
        self.curkey['uids'].append(args[9])
        self.uids.append(args[9])

    def handle_status(self, key, value):
        pass

class Crypt(Verify):
    "Handle status messages for --encrypt and --decrypt"
    def __init__(self, gpg):
        Verify.__init__(self, gpg)
        self.data = ''
        self.ok = False
        self.status = ''

    def __nonzero__(self):
        if self.ok: return True
        return False

    __bool__ = __nonzero__

    def __str__(self):
        return self.data.decode(self.gpg.encoding, self.gpg.decode_errors)

    def handle_status(self, key, value):
        if key in ("ENC_TO", "USERID_HINT", "GOODMDC", "END_DECRYPTION",
                   "BEGIN_SIGNING", "NO_SECKEY", "ERROR", "NODATA",
                   "CARDCTRL"):
            # in the case of ERROR, this is because a more specific error
            # message will have come first
            pass
        elif key in ("NEED_PASSPHRASE", "BAD_PASSPHRASE", "GOOD_PASSPHRASE",
                     "MISSING_PASSPHRASE", "DECRYPTION_FAILED",
                     "KEY_NOT_CREATED"):
            self.status = key.replace("_", " ").lower()
        elif key == "NEED_PASSPHRASE_SYM":
            self.status = 'need symmetric passphrase'
        elif key == "BEGIN_DECRYPTION":
            self.status = 'decryption incomplete'
        elif key == "BEGIN_ENCRYPTION":
            self.status = 'encryption incomplete'
        elif key == "DECRYPTION_OKAY":
            self.status = 'decryption ok'
            self.ok = True
        elif key == "END_ENCRYPTION":
            self.status = 'encryption ok'
            self.ok = True
        elif key == "INV_RECP":
            self.status = 'invalid recipient'
        elif key == "KEYEXPIRED":
            self.status = 'key expired'
        elif key == "SIG_CREATED":
            self.status = 'sig created'
        elif key == "SIGEXPIRED":
            self.status = 'sig expired'
        else:
            Verify.handle_status(self, key, value)

class GenKey(object):
    "Handle status messages for --gen-key"
    def __init__(self, gpg):
        self.gpg = gpg
        self.type = None
        self.fingerprint = None

    def __nonzero__(self):
        if self.fingerprint: return True
        return False

    __bool__ = __nonzero__

    def __str__(self):
        return self.fingerprint or ''

    def handle_status(self, key, value):
        if key in ("PROGRESS", "GOOD_PASSPHRASE", "NODATA"):
            pass
        elif key == "KEY_CREATED":
            (self.type,self.fingerprint) = value.split()
        else:
            raise ValueError("Unknown status message: %r" % key)

class DeleteResult(object):
    "Handle status messages for --delete-key and --delete-secret-key"
    def __init__(self, gpg):
        self.gpg = gpg
        self.status = 'ok'

    def __str__(self):
        return self.status

    problem_reason = {
        '1': 'No such key',
        '2': 'Must delete secret key first',
        '3': 'Ambigious specification',
    }

    def handle_status(self, key, value):
        if key == "DELETE_PROBLEM":
            self.status = self.problem_reason.get(value,
                                                  "Unknown error: %r" % value)
        else:
            raise ValueError("Unknown status message: %r" % key)

class Sign(object):
    "Handle status messages for --sign"
    def __init__(self, gpg):
        self.gpg = gpg
        self.type = None
        self.fingerprint = None

    def __nonzero__(self):
        return self.fingerprint is not None

    __bool__ = __nonzero__

    def __str__(self):
        return self.data.decode(self.gpg.encoding, self.gpg.decode_errors)

    def handle_status(self, key, value):
        if key in ("USERID_HINT", "NEED_PASSPHRASE", "BAD_PASSPHRASE",
                   "GOOD_PASSPHRASE", "BEGIN_SIGNING", "CARDCTRL"):
            pass
        elif key == "SIG_CREATED":
            (self.type,
             algo, hashalgo, cls,
             self.timestamp, self.fingerprint
             ) = value.split()
        else:
            raise ValueError("Unknown status message: %r" % key)


class GPG(object):

    decode_errors = 'strict'

    result_map = {
        'crypt': Crypt,
        'delete': DeleteResult,
        'generate': GenKey,
        'import': ImportResult,
        'list': ListKeys,
        'sign': Sign,
        'verify': Verify,
    }

    "Encapsulate access to the gpg executable"
    def __init__(self, gpgbinary='gpg', gnupghome=None, verbose=False,
                 use_agent=False, keyring=None):
        """Initialize a GPG process wrapper.  Options are:

        gpgbinary -- full pathname for GPG binary.

        gnupghome -- full pathname to where we can find the public and
        private keyrings.  Default is whatever gpg defaults to.
        keyring -- name of alternative keyring file to use. If specified,
        the default keyring is not used.
        """
        self.gpgbinary = gpgbinary
        self.gnupghome = gnupghome
        self.keyring = keyring
        self.verbose = verbose
        self.use_agent = use_agent
        self.encoding = locale.getpreferredencoding()
        if self.encoding is None: # This happens on Jython!
            self.encoding = sys.stdin.encoding
        if gnupghome and not os.path.isdir(self.gnupghome):
            os.makedirs(self.gnupghome,0x1C0)
        p = self._open_subprocess(["--version"])
        result = self.result_map['verify'](self) # any result will do for this
        self._collect_output(p, result, stdin=p.stdin)
        if p.returncode != 0:
            raise ValueError("Error invoking gpg: %s: %s" % (p.returncode,
                                                             result.stderr))

    def _open_subprocess(self, args, passphrase=False):
        # Internal method: open a pipe to a GPG subprocess and return
        # the file objects for communicating with it.
        cmd = [self.gpgbinary, '--status-fd 2 --no-tty']
        if self.gnupghome:
            cmd.append('--homedir "%s" ' % self.gnupghome)
        if self.keyring:
            cmd.append('--no-default-keyring --keyring "%s" ' % self.keyring)
        if passphrase:
            cmd.append('--batch --passphrase-fd 0')
        if self.use_agent:
            cmd.append('--use-agent')
        cmd.extend(args)
        cmd = ' '.join(cmd)
        if self.verbose:
            print(cmd)
        logger.debug("%s", cmd)
        return Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)

    def _read_response(self, stream, result):
        # Internal method: reads all the stderr output from GPG, taking notice
        # only of lines that begin with the magic [GNUPG:] prefix.
        #
        # Calls methods on the response object for each valid token found,
        # with the arg being the remainder of the status line.
        lines = []
        while True:
            line = stream.readline()
            if len(line) == 0:
                break
            lines.append(line)
            line = line.rstrip()
            if self.verbose:
                print(line)
            logger.debug("%s", line)
            if line[0:9] == '[GNUPG:] ':
                # Chop off the prefix
                line = line[9:]
                L = line.split(None, 1)
                keyword = L[0]
                if len(L) > 1:
                    value = L[1]
                else:
                    value = ""
                result.handle_status(keyword, value)
        result.stderr = ''.join(lines)

    def _read_data(self, stream, result):
        # Read the contents of the file from GPG's stdout
        chunks = []
        while True:
            data = stream.read(1024)
            if len(data) == 0:
                break
            logger.debug("chunk: %r" % data[:256])
            chunks.append(data)
        if _py3k:
            # Join using b'' or '', as appropriate
            result.data = type(data)().join(chunks)
        else:
            result.data = ''.join(chunks)

    def _collect_output(self, process, result, writer=None, stdin=None):
        """
        Drain the subprocesses output streams, writing the collected output
        to the result. If a writer thread (writing to the subprocess) is given,
        make sure it's joined before returning. If a stdin stream is given,
        close it before returning.
        """
        stderr = codecs.getreader(self.encoding)(process.stderr)
        rr = threading.Thread(target=self._read_response, args=(stderr, result))
        rr.setDaemon(True)
        logger.debug('stderr reader: %r', rr)
        rr.start()

        stdout = process.stdout
        dr = threading.Thread(target=self._read_data, args=(stdout, result))
        dr.setDaemon(True)
        logger.debug('stdout reader: %r', dr)
        dr.start()

        dr.join()
        rr.join()
        if writer is not None:
            writer.join()
        process.wait()
        if stdin is not None:
            try:
                stdin.close()
            except IOError:
                pass
        stderr.close()
        stdout.close()

    def _handle_io(self, args, file, result, passphrase=None, binary=False):
        "Handle a call to GPG - pass input data, collect output data"
        # Handle a basic data call - pass data to GPG, handle the output
        # including status information. Garbage In, Garbage Out :)
        p = self._open_subprocess(args, passphrase is not None)
        if not binary:
            stdin = codecs.getwriter(self.encoding)(p.stdin)
        else:
            stdin = p.stdin
        if passphrase:
            _write_passphrase(stdin, passphrase, self.encoding)
        writer = _threaded_copy_data(file, stdin)
        self._collect_output(p, result, writer, stdin)
        return result

    #
    # SIGNATURE METHODS
    #
    def sign(self, message, **kwargs):
        """sign message"""
        f = _make_binary_stream(message, self.encoding)
        result = self.sign_file(f, **kwargs)
        f.close()
        return result

    def sign_file(self, file, keyid=None, passphrase=None, clearsign=True,
                  detach=False, binary=False):
        """sign file"""
        logger.debug("sign_file: %s", file)
        if binary:
            args = ['-s']
        else:
            args = ['-sa']
        # You can't specify detach-sign and clearsign together: gpg ignores
        # the detach-sign in that case.
        if detach:
            args.append("--detach-sign")
        elif clearsign:
            args.append("--clearsign")
        if keyid:
            args.append('--default-key "%s"' % keyid)
        args.extend(['--no-version', "--comment ''"])
        result = self.result_map['sign'](self)
        #We could use _handle_io here except for the fact that if the
        #passphrase is bad, gpg bails and you can't write the message.
        p = self._open_subprocess(args, passphrase is not None)
        try:
            stdin = p.stdin
            if passphrase:
                _write_passphrase(stdin, passphrase, self.encoding)
            writer = _threaded_copy_data(file, stdin)
        except IOError:
            logging.exception("error writing message")
            writer = None
        self._collect_output(p, result, writer, stdin)
        return result

    def verify(self, data):
        """Verify the signature on the contents of the string 'data'

        >>> gpg = GPG(gnupghome="keys")
        >>> input = gpg.gen_key_input(Passphrase='foo')
        >>> key = gpg.gen_key(input)
        >>> assert key
        >>> sig = gpg.sign('hello',keyid=key.fingerprint,passphrase='bar')
        >>> assert not sig
        >>> sig = gpg.sign('hello',keyid=key.fingerprint,passphrase='foo')
        >>> assert sig
        >>> verify = gpg.verify(sig.data)
        >>> assert verify

        """
        f = _make_binary_stream(data, self.encoding)
        result = self.verify_file(f)
        f.close()
        return result

    def verify_file(self, file, data_filename=None):
        "Verify the signature on the contents of the file-like object 'file'"
        logger.debug('verify_file: %r, %r', file, data_filename)
        result = self.result_map['verify'](self)
        args = ['--verify']
        if data_filename is None:
            self._handle_io(args, file, result, binary=True)
        else:
            logger.debug('Handling detached verification')
            import tempfile
            fd, fn = tempfile.mkstemp(prefix='pygpg')
            s = file.read()
            file.close()
            logger.debug('Wrote to temp file: %r', s)
            os.write(fd, s)
            os.close(fd)
            args.append(fn)
            args.append('"%s"' % data_filename)
            try:
                p = self._open_subprocess(args)
                self._collect_output(p, result, stdin=p.stdin)
            finally:
                os.unlink(fn)
        return result

    #
    # KEY MANAGEMENT
    #

    def import_keys(self, key_data):
        """ import the key_data into our keyring

        >>> import shutil
        >>> shutil.rmtree("keys")
        >>> gpg = GPG(gnupghome="keys")
        >>> input = gpg.gen_key_input()
        >>> result = gpg.gen_key(input)
        >>> print1 = result.fingerprint
        >>> result = gpg.gen_key(input)
        >>> print2 = result.fingerprint
        >>> pubkey1 = gpg.export_keys(print1)
        >>> seckey1 = gpg.export_keys(print1,secret=True)
        >>> seckeys = gpg.list_keys(secret=True)
        >>> pubkeys = gpg.list_keys()
        >>> assert print1 in seckeys.fingerprints
        >>> assert print1 in pubkeys.fingerprints
        >>> str(gpg.delete_keys(print1))
        'Must delete secret key first'
        >>> str(gpg.delete_keys(print1,secret=True))
        'ok'
        >>> str(gpg.delete_keys(print1))
        'ok'
        >>> str(gpg.delete_keys("nosuchkey"))
        'No such key'
        >>> seckeys = gpg.list_keys(secret=True)
        >>> pubkeys = gpg.list_keys()
        >>> assert not print1 in seckeys.fingerprints
        >>> assert not print1 in pubkeys.fingerprints
        >>> result = gpg.import_keys('foo')
        >>> assert not result
        >>> result = gpg.import_keys(pubkey1)
        >>> pubkeys = gpg.list_keys()
        >>> seckeys = gpg.list_keys(secret=True)
        >>> assert not print1 in seckeys.fingerprints
        >>> assert print1 in pubkeys.fingerprints
        >>> result = gpg.import_keys(seckey1)
        >>> assert result
        >>> seckeys = gpg.list_keys(secret=True)
        >>> pubkeys = gpg.list_keys()
        >>> assert print1 in seckeys.fingerprints
        >>> assert print1 in pubkeys.fingerprints
        >>> assert print2 in pubkeys.fingerprints

        """
        result = self.result_map['import'](self)
        logger.debug('import_keys: %r', key_data[:256])
        data = _make_binary_stream(key_data, self.encoding)
        self._handle_io(['--import'], data, result, binary=True)
        logger.debug('import_keys result: %r', result.__dict__)
        data.close()
        return result

    def recv_keys(self, keyserver, *keyids):
        """Import a key from a keyserver

        >>> import shutil
        >>> shutil.rmtree("keys")
        >>> gpg = GPG(gnupghome="keys")
        >>> result = gpg.recv_keys('pgp.mit.edu', '3FF0DB166A7476EA')
        >>> assert result

        """
        result = self.result_map['import'](self)
        logger.debug('recv_keys: %r', keyids)
        data = _make_binary_stream("", self.encoding)
        #data = ""
        args = ['--keyserver', keyserver, '--recv-keys']
        args.extend(keyids)
        self._handle_io(args, data, result, binary=True)
        logger.debug('recv_keys result: %r', result.__dict__)
        data.close()
        return result

    def delete_keys(self, fingerprints, secret=False):
        which='key'
        if secret:
            which='secret-key'
        if _is_sequence(fingerprints):
            fingerprints = ' '.join(fingerprints)
        args = ['--batch --delete-%s "%s"' % (which, fingerprints)]
        result = self.result_map['delete'](self)
        p = self._open_subprocess(args)
        self._collect_output(p, result, stdin=p.stdin)
        return result

    def export_keys(self, keyids, secret=False):
        "export the indicated keys. 'keyid' is anything gpg accepts"
        which=''
        if secret:
            which='-secret-key'
        if _is_sequence(keyids):
            keyids = ' '.join(['"%s"' % k for k in keyids])
        args = ["--armor --export%s %s" % (which, keyids)]
        p = self._open_subprocess(args)
        # gpg --export produces no status-fd output; stdout will be
        # empty in case of failure
        #stdout, stderr = p.communicate()
        result = self.result_map['delete'](self) # any result will do
        self._collect_output(p, result, stdin=p.stdin)
        logger.debug('export_keys result: %r', result.data)
        return result.data.decode(self.encoding, self.decode_errors)

    def list_keys(self, secret=False):
        """ list the keys currently in the keyring

        >>> import shutil
        >>> shutil.rmtree("keys")
        >>> gpg = GPG(gnupghome="keys")
        >>> input = gpg.gen_key_input()
        >>> result = gpg.gen_key(input)
        >>> print1 = result.fingerprint
        >>> result = gpg.gen_key(input)
        >>> print2 = result.fingerprint
        >>> pubkeys = gpg.list_keys()
        >>> assert print1 in pubkeys.fingerprints
        >>> assert print2 in pubkeys.fingerprints

        """

        which='keys'
        if secret:
            which='secret-keys'
        args = "--list-%s --fixed-list-mode --fingerprint --with-colons" % (which,)
        args = [args]
        p = self._open_subprocess(args)

        # there might be some status thingumy here I should handle... (amk)
        # ...nope, unless you care about expired sigs or keys (stevegt)

        # Get the response information
        result = self.result_map['list'](self)
        self._collect_output(p, result, stdin=p.stdin)
        lines = result.data.decode(self.encoding,
                                   self.decode_errors).splitlines()
        valid_keywords = 'pub uid sec fpr'.split()
        for line in lines:
            if self.verbose:
                print(line)
            logger.debug("line: %r", line.rstrip())
            if not line:
                break
            L = line.strip().split(':')
            if not L:
                continue
            keyword = L[0]
            if keyword in valid_keywords:
                getattr(result, keyword)(L)
        return result

    def gen_key(self, input):
        """Generate a key; you might use gen_key_input() to create the
        control input.

        >>> gpg = GPG(gnupghome="keys")
        >>> input = gpg.gen_key_input()
        >>> result = gpg.gen_key(input)
        >>> assert result
        >>> result = gpg.gen_key('foo')
        >>> assert not result

        """
        args = ["--gen-key --batch"]
        result = self.result_map['generate'](self)
        f = _make_binary_stream(input, self.encoding)
        self._handle_io(args, f, result, binary=True)
        f.close()
        return result

    def gen_key_input(self, **kwargs):
        """
        Generate --gen-key input per gpg doc/DETAILS
        """
        parms = {}
        for key, val in list(kwargs.items()):
            key = key.replace('_','-').title()
            parms[key] = val
        parms.setdefault('Key-Type','RSA')
        parms.setdefault('Key-Length',1024)
        parms.setdefault('Name-Real', "Autogenerated Key")
        parms.setdefault('Name-Comment', "Generated by gnupg.py")
        try:
            logname = os.environ['LOGNAME']
        except KeyError:
            logname = os.environ['USERNAME']
        hostname = socket.gethostname()
        parms.setdefault('Name-Email', "%s@%s" % (logname.replace(' ', '_'),
                                                  hostname))
        out = "Key-Type: %s\n" % parms.pop('Key-Type')
        for key, val in list(parms.items()):
            out += "%s: %s\n" % (key, val)
        out += "%commit\n"
        return out

        # Key-Type: RSA
        # Key-Length: 1024
        # Name-Real: ISdlink Server on %s
        # Name-Comment: Created by %s
        # Name-Email: isdlink@%s
        # Expire-Date: 0
        # %commit
        #
        #
        # Key-Type: DSA
        # Key-Length: 1024
        # Subkey-Type: ELG-E
        # Subkey-Length: 1024
        # Name-Real: Joe Tester
        # Name-Comment: with stupid passphrase
        # Name-Email: joe@foo.bar
        # Expire-Date: 0
        # Passphrase: abc
        # %pubring foo.pub
        # %secring foo.sec
        # %commit

    #
    # ENCRYPTION
    #
    def encrypt_file(self, file, recipients, sign=None,
            always_trust=False, passphrase=None,
            armor=True, output=None, symmetric=False):
        "Encrypt the message read from the file-like object 'file'"
        args = ['--no-version', "--comment ''"]
        if symmetric:
            args.append('--symmetric')
        else:
            args.append('--encrypt')
            if not _is_sequence(recipients):
                recipients = (recipients,)
            for recipient in recipients:
                args.append('--recipient "%s"' % recipient)
        if armor:   # create ascii-armored output - set to False for binary output
            args.append('--armor')
        if output:  # write the output to a file with the specified name
            if os.path.exists(output):
                os.remove(output) # to avoid overwrite confirmation message
            args.append('--output "%s"' % output)
        if sign:
            args.append('--sign --default-key "%s"' % sign)
        if always_trust:
            args.append("--always-trust")
        result = self.result_map['crypt'](self)
        self._handle_io(args, file, result, passphrase=passphrase, binary=True)
        logger.debug('encrypt result: %r', result.data)
        return result

    def encrypt(self, data, recipients, **kwargs):
        """Encrypt the message contained in the string 'data'

        >>> import shutil
        >>> if os.path.exists("keys"):
        ...     shutil.rmtree("keys")
        >>> gpg = GPG(gnupghome="keys")
        >>> input = gpg.gen_key_input(passphrase='foo')
        >>> result = gpg.gen_key(input)
        >>> print1 = result.fingerprint
        >>> input = gpg.gen_key_input()
        >>> result = gpg.gen_key(input)
        >>> print2 = result.fingerprint
        >>> result = gpg.encrypt("hello",print2)
        >>> message = str(result)
        >>> assert message != 'hello'
        >>> result = gpg.decrypt(message)
        >>> assert result
        >>> str(result)
        'hello'
        >>> result = gpg.encrypt("hello again",print1)
        >>> message = str(result)
        >>> result = gpg.decrypt(message)
        >>> result.status == 'need passphrase'
        True
        >>> result = gpg.decrypt(message,passphrase='bar')
        >>> result.status in ('decryption failed', 'bad passphrase')
        True
        >>> assert not result
        >>> result = gpg.decrypt(message,passphrase='foo')
        >>> result.status == 'decryption ok'
        True
        >>> str(result)
        'hello again'
        >>> result = gpg.encrypt("signed hello",print2,sign=print1)
        >>> result.status == 'need passphrase'
        True
        >>> result = gpg.encrypt("signed hello",print2,sign=print1,passphrase='foo')
        >>> result.status == 'encryption ok'
        True
        >>> message = str(result)
        >>> result = gpg.decrypt(message)
        >>> result.status == 'decryption ok'
        True
        >>> assert result.fingerprint == print1

        """
        data = _make_binary_stream(data, self.encoding)
        result = self.encrypt_file(data, recipients, **kwargs)
        data.close()
        return result

    def decrypt(self, message, **kwargs):
        data = _make_binary_stream(message, self.encoding)
        result = self.decrypt_file(data, **kwargs)
        data.close()
        return result

    def decrypt_file(self, file, always_trust=False, passphrase=None,
                     output=None):
        args = ["--decrypt"]
        if output:  # write the output to a file with the specified name
            if os.path.exists(output):
                os.remove(output) # to avoid overwrite confirmation message
            args.append('--output "%s"' % output)
        if always_trust:
            args.append("--always-trust")
        result = self.result_map['crypt'](self)
        self._handle_io(args, file, result, passphrase, binary=True)
        logger.debug('decrypt result: %r', result.data)
        return result




import unittest
from slixmpp.test import SlixTest
from slixmpp.util import (
    MemoryCache, MemoryPerJidCache,
    FileSystemCache, FileSystemPerJidCache
)
from tempfile import TemporaryDirectory

class TestCacheClass(SlixTest):

    def testMemoryCache(self):
        cache = MemoryCache()

        cache.store("test", "test_value")
        self.assertEqual(cache.retrieve("test"), "test_value")
        self.assertEqual(cache.retrieve("test2"), None)

        cache.remove("test")
        self.assertEqual(cache.retrieve("test"), None)

    def testMemoryPerJidcache(self):
        cache = MemoryPerJidCache()

        cache.store_by_jid("test@example.com", "test", "test_value")
        self.assertEqual(
            cache.retrieve_by_jid("test@example.com", "test"),
            "test_value"
        )

        cache.remove_by_jid("test@example.com", "test")
        self.assertEqual(
            cache.retrieve_by_jid("test@example.com", "test"),
            None
        )

    def testFileSystemCache(self):
        def failing_decode(value):
            if value == "failme":
                raise Exception("you failed")
            return value
        with TemporaryDirectory() as tmpdir:
            cache = FileSystemCache(tmpdir, "test", decode=failing_decode)
            cache.store("test", "test_value")
            cache.store("test2", "failme")
            self.assertEqual(
                cache.retrieve("test"),
                "test_value"
            )
            cache.remove("test")
            self.assertEqual(
                cache.retrieve("test"),
                None
            )

            self.assertEqual(
                cache.retrieve("test2"),
                None
            )

    def testFileSystemPerJidCache(self):
        with TemporaryDirectory() as tmpdir:
            cache = FileSystemPerJidCache(tmpdir, "test")
            cache.store_by_jid("test@example.com", "test", "test_value")
            self.assertEqual(
                cache.retrieve_by_jid("test@example.com", "test"),
                "test_value"
            )
            cache.remove_by_jid("test@example.com", "test")
            self.assertEqual(
                cache.retrieve_by_jid("test@example.com", "test"),
                None
            )

suite = unittest.TestLoader().loadTestsFromTestCase(TestCacheClass)
